import asyncio
from functools import lru_cache
import uuid
import time
from typing import Optional, Callable, Any, Dict, Union, List

from PySide6.QtWidgets import QVBoxLayout, QWidget, QHBoxLayout, QPlainTextEdit, QLineEdit, QApplication, QSizePolicy, QPushButton, QFrame, QGroupBox, QLabel
from PySide6.QtCore import Qt, QPoint, Signal, QTimer, QEvent
from PySide6.QtGui import QCursor, QFontMetrics

from qtmui.hooks import State
from qtmui.utils.translator import getTranslatedText
from ..li import Li

from ..menu import Menu

from .flow_layout import FlowLayout
from .tags_view import TagsView

from qtmui.hooks import useState

from qtmui.material.styles import useTheme
from qtmui.i18n.use_translation import translate, i18n
from ..utils.functions import _get_fn_args

from ...qtmui_assets import QTMUI_ASSETS

from ..py_iconify import PyIconify


from qtmui.material.styles.create_theme.components.get_qss_styles import get_qss_style

from ..widget_base import PyWidgetBase

from ..utils.validate_params import _validate_param

class Select(QGroupBox):

    showPlaceholder = Signal()
    valueChanged = Signal(object)
    updatePopupPosition = Signal()
    createMenu = Signal()
    renderSearchMenuItems = Signal(object)

    VALID_COLORS = ['primary', 'secondary', 'error', 'info', 'success', 'warning']
    VALID_MARGINS = ['dense', 'none', 'normal']
    VALID_SIZES = ['medium', 'small']
    VALID_VARIANTS = ['filled', 'outlined', 'standard']

    def __init__(
        self,
        key: str= None,
        autoWidth: Optional[Union[State, bool]] = False, # chỉ dùng cho Select
        color: Optional[Union[State, str]] = "primary",
        defaultOpen: Optional[Union[State, bool]] = False, # chỉ dùng cho Select
        defaultValue: Optional[Union[State, object]] = None,
        disabled: Optional[Union[State, bool]] = False,
        children: Optional[Union[State, list]] = None,
        error: Optional[Union[State, bool]] = False,
        fullWidth: Optional[Union[State, bool]] = False,
        helperText: Optional[Union[State, str]] = None,
        id: Optional[Union[State, str]] = None,
        label: Optional[Union[State, str, Callable]] = None,
        multiple: Optional[Union[State, bool]] = False,
        menuHeight: Optional[Union[State, int]] = 400,
        name: Optional[Union[State, str]] = None,
        getOptionLabel: Optional[Union[State, Callable]] = None,
        open: Optional[Union[State, bool]] = False, # chỉ dùng cho Select
        onOpen: Optional[Union[State, Callable]] = None, # chỉ dùng cho Select
        onClose: Optional[Union[State, Callable]] = None, # chỉ dùng cho Select
        onChange: Optional[Union[State, Callable]] = None,
        options: Optional[Union[State, object]] = None,
        hiddenLabel: Optional[Union[State, bool]] = False,
        placeholder: Optional[Union[State, str]] = None,
        required: Optional[Union[State, bool]] = False,
        size: Optional[Union[State, str]] = "medium",
        sx: Optional[Union[State, Dict, Callable, str]] = None,
        step: Optional[Union[State, int, float]] = None,
        renderSelectOptions: Optional[Union[State, Callable]] = None,
        renderOption: Optional[Union[State, Callable]] = None,
        renderValue: Optional[Union[State, Callable]] = None,
        renderTags: Optional[Union[State, Callable]] = None,
        readOnly: Optional[Union[State, bool]] = False, # chỉ dùng cho Select
        value = None,
        variant: Optional[Union[State, str]] = "outlined",
        *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.setObjectName(str(uuid.uuid4()))
        start_time = time.time()

        self._key = key
        
        if self._key == "proxyTypekyyyyyyyyyyyyyyyyy":
            print('ProfileTextField__________________', self._key)

        self._key = key
        self._color = color
        self._defaultValue = defaultValue
        self._disabled = disabled
        self._error = error
        self._fullWidth = fullWidth
        self._helperText = helperText
        self._id = id
        self._label = label
        self._multiple = multiple
        self._menuHeight = menuHeight
        self._name = name
        self._getOptionLabel = getOptionLabel
        self._onChange = onChange
        self._options = options
        self._hiddenLabel = hiddenLabel
        self._placeholder = placeholder
        self._required = required
        self._size = size
        self._sx = sx
        self._step = step
        self._renderSelectOptions = renderSelectOptions
        self._renderOption = renderOption
        self._renderValue = renderValue
        self._renderTags = renderTags
        self._value = value
        self._variant = variant

        self.__init_ui()
        
        end_time = time.time()
        
        # print("time_loading_textfield___________________________", (end_time - start_time)/0.00013971328735351562)


    def _connect_signals_props(self):
        if isinstance(self._value, State):
            self._value.valueChanged.connect(self._set_data)

    def __init_ui(self):
        self._init_done = False

        self._connect_signals_props()

        self.theme = useTheme()
        
        i18n.langChanged.connect(self._set_translated_label)

        if self._fullWidth:
            if not self._sx:
                self._sx = {}
            self._sx.update({"width": "100%"})

        self._data = None

        self._tagsWidget = None
        self._selected_options = {}
        self._inputFieldHasFocusIn = False


        self._inputField = None
        self._popup = None
        self._btnSelect = None

        self._menu = None
        self._menuItems, self._setMenuItems = useState([])
        
        self.setLayout(QHBoxLayout())
        self.layout().setContentsMargins(0,0,0,0)
        
        self._focused, self._setFocused = useState(False)
        self._hovered, self._setHovered = useState(False)

        if self._multiple and self._renderTags:
            self._inputField = QPlainTextEdit()
            self._inputField.mousePressEvent = self.onInputFieldMousePressEvent
            self._inputField.focusOutEvent = self.onInputFieldFocusOutEvent
            self._inputField.setStyleSheet(
                f"""
                QPlainTextEdit {{
                    border: none;
                    width: 100%;
                    min-height: 22px;
                    background-color: transparent;
                    padding-left: 5px;
                }}
                """
            )
        else:
            self._inputField = QLineEdit()
            self._inputField.mousePressEvent = self.onInputFieldMousePressEvent
            self._inputField.focusOutEvent = self.onInputFieldFocusOutEvent
            self._inputField.setStyleSheet(
                f"""
                QLineEdit {{
                    border: none;
                    width: 100%;
                    background-color: transparent;
                    padding-left: 8px;
                }}
                """
            )
            
        if not self._defaultValue:
            self._inputField.setPlaceholderText(getTranslatedText(self._label))

        self.setMinimumHeight(44)
        
        self._inputField.setReadOnly(True)
        self._inputField.setCursor(Qt.PointingHandCursor)
        
        self.layout().addWidget(self._inputField)

        if self._defaultValue is not None:
            self._set_data(self._defaultValue, setText=True)

        if self._value is not None:
            self._set_data( self._value, setText=True)

        if self._disabled:
            self.setEnabled(False)

        PyWidgetBase._installTooltipFilter(self)


        # ASYN INIT
        QTimer.singleShot(0, self._scheduleSetupSelectMode)
        
        # chỗ này nghiên cứu lại kỹ hơn, chưa kiểm định lỗi, ít khi menu dài tới hàm trăm phần tử, nên k nhất thiết load trước
        # QTimer.singleShot(500, self._scheduleCraeteMenu) 

        self.theme.state.valueChanged.connect(self._onThemeChanged)
        # QTimer.singleShot(0, self._scheduleSetStyleSheet)
        self._set_stylesheet()
        
        if self._multiple:
            QTimer.singleShot(0, self._scheduleSetupDefaultTags)
            
        self.destroyed.connect(lambda obj: self._onDestroy())
        
        self._init_done = True

    def _onDestroy(self, obj=None):
        # Cancel task nếu đang chạy
        if hasattr(self, "_setupDefaultTags") and self._setupDefaultTags and not self._setupDefaultTags.done():
            self._setupDefaultTags.cancel()
        if hasattr(self, "_setupStyleSheet") and self._setupStyleSheet and not self._setupStyleSheet.done():
            self._setupStyleSheet.cancel()
        if hasattr(self, "_setupSelectMode") and self._setupSelectMode and not self._setupSelectMode.done():
            self._setupSelectMode.cancel()
        if hasattr(self, "_setupMenu") and self._setupMenu and not self._setupMenu.done():
            self._setupMenu.cancel()
            
    def _onThemeChanged(self):
        if not self.isVisible():
            return
        QTimer.singleShot(0, self._scheduleSetStyleSheet)
            
    def _scheduleSetupDefaultTags(self):
        self._setupDefaultTags = asyncio.ensure_future(self._lazySetupDefautTags())

    async def _lazySetupDefautTags(self):
        if self._defaultValue:
            for option in self._defaultValue:
                key = self._getOptionLabel(option)
                self._selected_options.update({key: option})
            self.tags = self._renderTags(self._selected_options, self.getTagProps)
            self._setFlowLayout(FlowLayout(self, children=self.tags))
            
    def _scheduleSetStyleSheet(self):
        self._setupStyleSheet = asyncio.ensure_future(self._lazy_set_stylesheet())

    async def _lazy_set_stylesheet(self):
        self._set_stylesheet()
        
    def _scheduleSetupSelectMode(self):
        self._setupSelectMode = asyncio.ensure_future(self._lazy_setupSelectMode())

    async def _lazy_setupSelectMode(self):
        self._setup_select_mode()

    def _scheduleCraeteMenu(self):
        self._setupMenu = asyncio.ensure_future(self._lazy_create_menu())

    async def _lazy_create_menu(self):
        self._create_menu()

    def _setup_select_mode(self):
        # self.updatePopupPosition.connect(self._update_popup_position) => xem lai

        # đặt ở đây để có _selectedKeys truyền cho QLineEditMultiple => setup placeholder đúng
        if self._options:
            self._selectedKeys, self._setSelectedKeys = useState(self._options[0] if isinstance(self._options[0], str) else self._options[0].get("value") or "")

            if self._multiple:
                if self._defaultValue:
                    selectedKeys = ["has_defaultValue"] # để thiết lập placeholder cho đúng
                    self._selectedKeys, self._setSelectedKeys = useState(selectedKeys)
            else:
                self._selectedKeys, self._setSelectedKeys = useState(
                    self._options[0] if isinstance(self._options[0], str) 
                    else self._getOptionLabel(self._options[0]) if self._getOptionLabel else self._options[0].get("value") or ""
                )
        elif self._renderOption: # cho Select - select thi k render tag ma render text value
            self._selectedKeys, self._setSelectedKeys = useState(self._data)

        elif self._selectOptions: 
            self._selectedKeys, self._setSelectedKeys = useState(self._data)

        self.action_frame = QWidget(self)
        self.action_frame.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        hlo_action_frame = QHBoxLayout(self.action_frame)
        hlo_action_frame.setContentsMargins(0,0,0,0)
        self.action_frame.setFixedWidth(54)
        self.action_frame.layout().setAlignment(Qt.AlignRight | Qt.AlignVCenter)

        if self._multiple:
            if self._renderTags:
                self._flowLayout, self._setFlowLayout = useState(FlowLayout(self, []))
                self._flow_layout_view = TagsView(content=self._flowLayout)
                self._flow_layout_view.layout().setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
                self._inputField.setReadOnly(True)
                self._inputField.setLayout(QHBoxLayout())
                self._inputField.layout().setContentsMargins(0,0,54,0)
                self._inputField.layout().addWidget(self._flow_layout_view)
                self._flow_layout_view.mousePressEvent = self.onFlowlayoutMousePressEvent
            elif self._renderValue:
                if self._data:
                    # self._inputField.setPlainText(self._renderValue(self._data))
                    self.setElidedText(self._inputField, self._renderValue(self._data))
                    

            if isinstance(self._renderOption, State):
                self._renderOption.valueChanged.connect(self._create_menu)


        if self._options and not self._getOptionLabel:
            raise AttributeError("Opp!! 'getOptionLabel' is required attribute!")
        
        self._btn_clear_text= QPushButton()
        self._btn_clear_text.setFixedSize(24, 24)
        self._btn_clear_text.setIcon(PyIconify(key=QTMUI_ASSETS.ICONS.CLOSE))
        self._btn_clear_text.setCursor(Qt.PointingHandCursor)
        self._btn_clear_text.clicked.connect(self._on_clear_text)
        self._btn_clear_text.hide()
        hlo_action_frame.addWidget(self._btn_clear_text)
        self._btn_clear_text.hide()
        self._btn_clear_text.setFocusPolicy(Qt.NoFocus)
        self._btn_clear_text.setAttribute(Qt.WA_ShowWithoutActivating, True)

        self._btnSelect = QPushButton()
        self._btnSelect.setCursor(Qt.PointingHandCursor)
        self._btnSelect.setFixedSize(24, 24)
            
        self._btnSelect.setIcon(PyIconify(key=QTMUI_ASSETS.ICONS.ARROW_DOWN))
        self._btnSelect.clicked.connect(self._toggle_popup)
        hlo_action_frame.addWidget(self._btnSelect)

        self._inputField.installEventFilter(self)
        
        self.setup_action_frame_and_adornment()
        

    # @timer
    def _set_stylesheet(self, component_styled=None):
        _theme_mode = useTheme().palette.mode
        self.setProperty("variant", self._variant)
        if self._renderTags:
            _renderTags = "True"
        else:
            _renderTags = "False"

        stylesheet = self._get_stylesheet(self._variant, self._size, _theme_mode, str(self._multiple),  _renderTags )
        stylesheet = stylesheet.replace("_________object_name_______", self.objectName())
        self.setStyleSheet(stylesheet)

    @classmethod
    @lru_cache(maxsize=128)
    def _get_stylesheet(cls, _variant: str, _size: str, _theme_mode: str, _multiple: str,  _renderTags:str):
        """Set the stylesheet for the Box."""
        theme = useTheme()

        # Chung cho toàn bộ input
        MuiInput = theme.components["MuiInput"].get("styles")
        MuiInput_root = MuiInput.get("root")
        MuiInput_root_qss = get_qss_style(MuiInput_root)
        MuiInput_title = MuiInput.get("title")
        MuiInput_title_qss = get_qss_style(MuiInput_title)
        MuiInput_title_slot_focused_qss = get_qss_style(MuiInput_title["slots"]["focused"])
        MuiInput_title_slot_error_qss = get_qss_style(MuiInput_title["slots"]["error"])
        MuiInput_title_slot_disabled_qss = get_qss_style(MuiInput_title["slots"]["disabled"])
        MuiInput_inputField = MuiInput.get("inputField")
        MuiInput_inputField_qss = get_qss_style(MuiInput_inputField)
        MuiInput_inputField_prop_multiline_qss = get_qss_style(MuiInput_inputField["props"]["multiline"][f"{_size}"])
        MuiInput_inputField_prop_multiple_qss = get_qss_style(MuiInput_inputField["props"]["multiple"])

        MuiInputSize = theme.components["MuiInputSize"]["styles"][f"{_size}"]
        MuiInputSize_qss = get_qss_style(MuiInputSize)

        MuiInputSize_not_multiple_qss = get_qss_style(MuiInputSize["no-multiple"])
        MuiInputSize_no_multiple_height = MuiInputSize["no-multiple"]["height"]

        # thiết lập tuỳ theo variant
        PyVariantInput = theme.components[f"Mui{_variant.capitalize()}Input"].get("styles")
        PyVariantInput_root = PyVariantInput.get("root")
        PyVariantInput_root_qss = get_qss_style(PyVariantInput_root)
        PyVariantInput_root_slot_hovered_qss = get_qss_style(PyVariantInput_root["slots"]["hover"])
        PyVariantInput_root_slot_focused_qss = get_qss_style(PyVariantInput_root["slots"]["focus"])
        
        PyVariantInput_inputField_root = PyVariantInput.get("inputField")["root"]
        PyVariantInput_inputField_root_qss = get_qss_style(PyVariantInput_inputField_root)

        PyVariantInput_root_prop_hasStartAdornment_qss = ""
        if PyVariantInput_inputField_root.get("props"):
            PyVariantInput_root_prop_hasStartAdornment_qss = get_qss_style(PyVariantInput_inputField_root["props"]["hasStartAdornment"])

        MuiInputSize = theme.components["MuiInputSize"]["styles"][f"{_size}"]
        MuiInputSize_qss = get_qss_style(MuiInputSize)

        # print('MuiInputSize_qss______________', MuiInputSize_qss)

        _________object_name_______ = "_________object_name_______"


        input_base_stylesheet = f"""
                #{_________object_name_______} {{
                    {MuiInput_root_qss}
                    {MuiInputSize_qss}
                }}

                #{_________object_name_______} QLineEdit, QPlainTextEdit {{
                    {MuiInput_inputField_qss}
                }}
                #{_________object_name_______}[multipleHasValue=true] QPlainTextEdit {{
                    {MuiInput_inputField_prop_multiple_qss}
                }}

                #{_________object_name_______} QPlainTextEdit{{
                    min-height: 40px;
                    padding-top: 6px;
                }}
                #{_________object_name_______}::title {{
                    {MuiInput_title_qss}
                }}
                #{_________object_name_______}[focused=true]::title {{
                    {MuiInput_title_slot_focused_qss}
                }}
                #{_________object_name_______}[error=true]::title {{
                    {MuiInput_title_slot_error_qss}
                }}
                #{_________object_name_______}[disabled=true]::title {{
                    {MuiInput_title_slot_disabled_qss}
                }}
            """

        _variant_stylesheet = f"""
                #{_________object_name_______}[variant={_variant}] {{
                    {PyVariantInput_root_qss}
                    {MuiInputSize_qss}
                }}
                #{_________object_name_______}[variant={_variant}]  QLineEdit, QPlainTextEdit, QTextEdit {{
                    {PyVariantInput_inputField_root_qss}
                }}
                #{_________object_name_______}[outlinedHasStartAdornment=true]  QLineEdit, QPlainTextEdit, QTextEdit {{
                    {PyVariantInput_root_prop_hasStartAdornment_qss}
                }}
                #{_________object_name_______}[hovered=true] {{
                    {PyVariantInput_root_slot_hovered_qss}
                }}
                #{_________object_name_______}[focused=true] {{
                    {PyVariantInput_root_slot_focused_qss}
                }}
            """
        
        return input_base_stylesheet + _variant_stylesheet


    def _create_menu(self):
        print("_create_menu called")

        try:
            if not self._menu:
                self._menu = Menu(
                    maxHeight=self._menuHeight,
                    children=self._menuItems
                )

            if self._options:
                self._selectOptions = []
                for _option in self._options:
                    props = {}
                    option = self._renderOption(props, _option)

                    if isinstance(option, Li):
                        option._set_selected_keys(self._selectedKeys)
                        option._set_on_changed(self._selected_change)
                    self._selectOptions.append(option)

            elif self._renderOption and isinstance(self._renderOption, Callable):
                self._selectOptions = self._renderOption(**(_get_fn_args(self._renderOption)))
                for option in self._selectOptions:
                    if isinstance(option, Li):
                        option._set_selected_keys(self._selectedKeys)
                        option._set_on_changed(self._selected_change)

            self._setMenuItems(self._selectOptions)
        except Exception as e:
            print(f"Error creating menu: {e}")

    def onFlowlayoutMousePressEvent(self, event=None):
        self._toggle_popup()
        

    def onChange(self, value=None):
        # nếu là select thì chưa set_value
        self._set_data(value, setText=False)


    def _set_data(self, value=None, valueChanged=True, setText=True):
        
        _selected_options = self._selected_options.copy()
        
        for key, val in _selected_options.items():
            if isinstance(value, list) and val not in value:
                del self._selected_options[key]
                
        if isinstance(value, State):
            self._data = value.value
        else:
            self._data = value

        if self._multiple:
            if self._data is not None:
                self._set_multiple_has_value_prop(True)
            else:
                self._set_multiple_has_value_prop(False)
        else:
            if self._data is not None:
                if self._variant == "filled":
                    self._set_filled_has_value_prop(True) # làm cho inputField padding top 1 khoảng cho giống mui JS
            else:
                if self._variant == "filled":
                    self._set_filled_has_value_prop(False) # làm cho inputField padding top 1 khoảng cho giống mui JS

        if value is not None: # value == "" cũng vào đây
            if self._data != "":
                self._show_label()

        if valueChanged:
            self.valueChanged.emit(self._data)

        if setText:
            self._set_text()

    def _set_text(self): # value
        if not self._multiple: # multiple thì set tag không cần set text
            if isinstance(self._data, dict):
                if self._getOptionLabel:
                    text = self._getOptionLabel(self._data)
                    self._inputField.setText(str(text))
            else:
                if self._getOptionLabel:
                    for option in self._options:
                        if isinstance(option, dict):
                            if self._getOptionLabel(option) == self._data: # options khong phai dạng value label
                                text = self._getOptionLabel(option)
                                self._inputField.setText(str(text))
                                break
                        elif isinstance(option, str):
                            if option == self._data:
                                text = self._getOptionLabel(option)
                                self._inputField.setText(str(text))
                                break
                else:
                    self._inputField.setText(str(self._data))

        else: # multiple
            if self._renderValue:
                self.setElidedText(self._inputField, self._renderValue(self._data))
            else:
                if not self._renderTags:
                    self._inputField.setText(self._renderValue(self._data))

    def setElidedText(self, line_edit: QLineEdit, text: str):
        """Đặt text vừa khít với chiều rộng QLineEdit (có ... nếu tràn)."""
        if not isinstance(line_edit, QLineEdit):
            raise TypeError("line_edit phải là QLineEdit")
        
        metrics = QFontMetrics(line_edit.font())
        available_width = line_edit.width() - 62  # trừ padding nhỏ bên trong và frame right (nút mũi tên)
        
        # Tạo text có dấu ... nếu text dài hơn
        elided = metrics.elidedText(text, Qt.TextElideMode.ElideRight, available_width)
        
        print('elided_____________', elided)
        
        line_edit.setText(elided)
        line_edit.setToolTip(text)  # gợi ý toàn bộ text khi hover (rất tiện)

    def _set_slot(self, data):
        if data.get("slot") == "error":
            self._inputField.setProperty("p-error", "true")
        elif data.get("slot") == "valid":
            self._inputField.setProperty("p-error", "false")

    def _update_popup_position(self):
        if hasattr(self, "_popup") and self._popup is not None:
            popup_position = self.mapToGlobal(QPoint(0, self.height()))
            self._popup.setGeometry(popup_position.x(), popup_position.y(), self.width(), 50)


    def _hide_popup(self):
        if self._popup and self._popup.isVisible():
            self._popup.setVisible(False)
        icon = PyIconify(key=QTMUI_ASSETS.ICONS.ARROW_DOWN)

        if self._btnSelect:
                self._btnSelect.setIcon(icon)
        self._popupShowed = False

    def _on_clear_text(self, e=None):
        self._onClearText = True

        if self._multiple:
            self._setSelectedKeys([])
            self._btn_clear_text.hide()
            self._selected_options = {}
            self._set_data([])
            self._setFlowLayout(FlowLayout(self, children=[]))
        else: # text, email, password
            self._set_data("")
            self._inputField.setText("")
            
        if self._placeholder:
            self._inputField.setPlaceholderText(getTranslatedText(self._placeholder))
        else:
            if self._label:
                self._inputField.setPlaceholderText("")
            else:
                self._inputField.setPlaceholderText("None")
            
        self._btn_clear_text.hide()
        
        if self._onChange:
            self._onChange(self._data)



        

    def _on_menu_item_selected(self, key):
        self._hide_popup()
        self._data = key


    def getTagProps(self, index):
        return {"onDelete": lambda key=index: self._on_tag_deleted(key)}

    def _on_tag_deleted(self, index):
        self._selected_change(index)

    def _update_selected_options(self, key=None, option=None):
        """
        Hàm này nhằm cập nhật giá trị self._selected_options cho Select
        """
        if key not in list(self._selected_options.values()):
            self._selected_options.update({key: key})
        else:
            self._selected_options.pop(key)

    def _setup_tags(self, key=None, option=None):
        if key:
            for _option in self._options:
                if self._getOptionLabel(_option) == key:
                    option = _option
                    break
        else:
            if option:
                if isinstance(option, dict):
                    key = self._getOptionLabel(option)

        if option not in self._selected_options.values():
            try:
                self._selected_options.update({key: option})
                self.tags = self._renderTags(self._selected_options, self.getTagProps)
            except Exception as e:
                import traceback
                traceback.print_exc()
            self._setFlowLayout(FlowLayout(self, children=self.tags))
        else:
            self._selected_options.pop(key)
            self.tags = self._renderTags(self._selected_options, self.getTagProps)
            self._setFlowLayout(FlowLayout(self, children=self.tags))

    def _selected_change(self, data):

        if isinstance(data, dict):
            key = data.get("value")
            value = data.get("value")
        else:
            key = data
            value = data

        if self._multiple:
            # for delete option item
            if key in self._selectedKeys.value:
                selected_keys = self._selectedKeys.value.copy()
                selected_keys.remove(key)
                self._setSelectedKeys(selected_keys)
                if self._renderTags:
                    self._setup_tags(key=key)
                elif self._renderValue:
                    self._update_selected_options(key)
                values = []
                for selected_option in self._selected_options.values():
                    values.append(selected_option)

                self._set_data(values)

                if self._onChange:
                    self._onChange(self._data)

                if self._renderValue:
                    # self._inputField.setText(self._renderValue(self._data))
                    self.setElidedText(self._inputField, self._renderValue(self._data))

                return
            
            if self._renderTags: # cho Autocomplete - render tag
                self._setup_tags(key=key, option=key)
            else: # cho Select
                self._update_selected_options(key)
            
            keys = []
            values = []
            for selected_option in self._selected_options.values():
                if isinstance(selected_option, dict):
                    values.append(selected_option)
                    if self._getOptionLabel:
                        key = self._getOptionLabel(selected_option)
                        keys.append(key)
                elif isinstance(selected_option, str):
                    values.append(selected_option)
                    keys.append(selected_option)

            self._setSelectedKeys(keys)
            self._set_data(values)
            
            if len(values) and self._renderTags:
                self._inputField.setPlaceholderText("")

            if self._onChange:
                self._onChange(self._data)

            if self._renderValue:
                self.setElidedText(self._inputField, self._renderValue(self._data))
                
            self._hide_popup()
        else: # not multiple
            self._setSelectedKeys(key)
            self._set_data(value)
            self._selected_options.update({key: key})                
            self._hide_popup()

            # Call onChange with the new value
            if self._onChange:
                # self._onChange(data) # contronled state mode không đúng, nhận vào list mà onchange lại là dict
                self._onChange(self._data)
                
        if self._menu:
            for item in self._menu.findChildren(Li):
                item.show()

    def _show_label(self):
        if not self._hiddenLabel:
            self._set_translated_label()

    def _set_translated_label(self):
        if not self._hiddenLabel:
            self.setTitle(getTranslatedText(self._label))

    def _hide_label(self):
        print('_hide_label')
        self.setTitle("")

    def _set_hovered_prop(self):
        if self._disabled:
            return
        if isinstance(self._hovered, State):
            self.setProperty("hovered", self._hovered.value)
        elif isinstance(self._hovered, bool):
            self.setProperty("hovered", self._hovered)
        
    def _set_focused_prop(self):
        if isinstance(self._focused, State):
            self.setProperty("focused", self._focused.value)
        elif isinstance(self._focused, bool):
            self.setProperty("focused", self._focused)
        
    def _set_multiple_has_value_prop(self, state):
        self.setProperty("multipleHasValue", state)
        
    def _set_filled_has_value_prop(self, state:bool):
        self.setProperty("filledHasValue", state)

    def _set_filled_foucusin_prop(self, state:bool):
        self.setProperty("filledHasValue", state)

    def _set_prop_has_start_adornment(self, state:bool):
        self.setProperty("hasStartAdornment", state)

    def setup_action_frame_and_adornment(self):
        if hasattr(self, 'action_frame'):
            self.action_frame.move(self.width() - self.action_frame.width(), (self.height() + 6 - self.action_frame.height()) // 2)
            if not self.action_frame.isVisible():
                self.action_frame.show()

    def _toggle_popup(self, e=None):
        # print(self._popup.isVisible())
        if self._popup and self._popup.isVisible():
            print("_toggle_popup>>>>>>>>>>hide popup")
            self._hide_popup()
        else:
            print("_toggle_popup>>>>>>>>>>show popup", self._popup.isVisible() if self._popup else "None")
            self._show_popup()

    def _show_popup(self, search=None):
        
        if not self._menu:
            self._create_menu()
        
        if self._popup and self._popupShowed:
            return

        if not self._popup:
            self._popup = QFrame(self)
            self._popup.setFocusPolicy(Qt.NoFocus)
            self._popup.setAttribute(Qt.WA_ShowWithoutActivating, True)
            self._popup.setWindowFlags(Qt.Window |  Qt.FramelessWindowHint | Qt.NoDropShadowWindowHint)
            self._popup.setAttribute(Qt.WA_TranslucentBackground, True) # Enable translucent background => setStyleSheet ko có tác dụng gì
            QApplication.instance().installEventFilter(self)
            self._popup.setObjectName(str(uuid.uuid4()))
            self._popup.setLayout(QVBoxLayout())
            self._popup.layout().setContentsMargins(0,0,9,9)
            self._popup.layout().addWidget(self._menu)
            self._popup.setAutoFillBackground(False)
            self._popup.setStyleSheet(f"#{self._popup.objectName()} {{background: transparent;border: 2px solid red;}}")

        # Lấy điểm cục bộ (x, y) của relativeTo và chuyển sang global
        global_point = self.mapToGlobal(QPoint(0, self.height()))
        
        # tính toán vị trí hiển thị popup
        main_window_pos_y = QApplication.instance().mainWindow.height()
        self_pos_y = global_point.y()
        
        self._popup.move(-3000, -3000) # di chuyển tạm để lấy kích thước
        self._popup.setVisible(True)

        if (main_window_pos_y - self_pos_y) < self._menu.height():
            global_point.setY(global_point.y() - self.height() - self._menu.height())
        
        self._popup.move(global_point)
        self._popup.setMinimumWidth(self.width())

        self._menu.setVisible(True)

        icon = PyIconify(key=QTMUI_ASSETS.ICONS.ARROW_UP)
        self._btnSelect.setIcon(icon)
        self._popupShowed = True

    def leaveEvent(self, event):
        if not self.underMouse() and not self._inputField.hasFocus():
            if hasattr(self, "_btn_clear_text") and self._btn_clear_text.isVisible():
                self._btn_clear_text.setVisible(False)
        return super().leaveEvent(event)

    def resizeEvent(self, event):
        self.setup_action_frame_and_adornment()
        return super().resizeEvent(event)
    
        
    def focusInEvent(self, event):
        self._setFocused(True)
        self._show_label()
        if self._variant == "filled":
            self._set_filled_foucusin_prop(True)
        return super().focusInEvent(event)
        
    def focusOutEvent(self, event):
        print('focusOutEvent___________')
        
        # dành cho thiết lập khi nhấn vào nút clear hay select
        if hasattr(self, '_btn_clear_text') and QApplication.widgetAt(QCursor.pos()) is self._btn_clear_text:
            try:
                QTimer.singleShot(0, lambda: self._inputField.setFocus(Qt.OtherFocusReason))
            except Exception as e:
                print(f"Error setting focus: {e}")
            return

        if hasattr(self, '_btnSelect') and QApplication.widgetAt(QCursor.pos()) is self._btnSelect:
            try:
                QTimer.singleShot(0, lambda: self._inputField.setFocus(Qt.OtherFocusReason))
            except Exception as e:
                print(f"Error setting focus: {e}")
            return
        
        self._setFocused(False)
        

        # nếu inputValue hiện tại không nằm trong options thì clear nó đi
        if self._multiple:
            if len(self._selectedKeys.value) == 0:
                self._hide_label()
                self._inputField.setPlaceholderText(getTranslatedText(self._label))
        else:
            if isinstance(self._inputField, QLineEdit):
                current_input_value = self._inputField.text()
            elif isinstance(self._inputField, QPlainTextEdit):
                current_input_value = self._inputField.toPlainText()
                
            option_values = []
            options = self._options
            if options:
                for option in options:
                    label = option
                    if self._getOptionLabel:
                        label = self._getOptionLabel(option)
                    option_values.append(label.lower())
            if current_input_value.lower() not in option_values:
                self._set_data(None)
                if isinstance(self._inputField, QLineEdit):
                    self._inputField.setText("")
                elif isinstance(self._inputField, QPlainTextEdit):
                    self._inputField.setPlainText("")
                self._inputField.setPlaceholderText(getTranslatedText(self._label))
            else:
                # set lại data dựa trên inputValue
                for option in options:
                    label = option
                    if self._getOptionLabel:
                        label = self._getOptionLabel(option)
                    if label.lower() == current_input_value.lower():
                        self._set_data(option)
                        self._inputField.setText(label)
                        break

        if self._data is None:
            if self._variant == "filled":
                self._set_filled_foucusin_prop(False)
            self._hide_label()

        if not self.underMouse():
            if hasattr(self, "_btn_clear_text") and self._btn_clear_text.isVisible():
                self._btn_clear_text.setVisible(False)
        return super().focusOutEvent(event)

    def enterEvent(self, event):
        self._setHovered(True)
        if self._data:
            if hasattr(self, "_btn_clear_text") and not self._btn_clear_text.isVisible():
                self._btn_clear_text.setVisible(True)
        return super().enterEvent(event)

    def leaveEvent(self, event):
        self._setHovered(False)
        return super().leaveEvent(event)

    def onInputFieldFocusOutEvent(self, event):
        print('onInputFieldFocusOutEvent')
        self.focusOutEvent(event)
        # return super().focusOutEvent(event)

    def onInputFieldMousePressEvent(self, event):
        self._toggle_popup()
        return super().mousePressEvent(event)
    
    def eventFilter(self, obj, event):
        if self._popup and self._popup.isVisible():
            if event.type() == QEvent.MouseButtonPress and not self.underMouse():
                if not self._popup.geometry().contains(event.globalPos()):
                    self._hide_popup()
            elif event.type() == QEvent.Wheel and ((not self.underMouse() and not self._popup.underMouse()) or obj == self._inputField):
                self._hide_popup()
            
        return super().eventFilter(obj, event)




import sys
import math
from dataclasses import dataclass, field
from typing import Dict, List, Union, Callable, Optional, Any, Tuple

from PySide6.QtCore import QObject, Property, QVariantAnimation, QEasingCurve
from PySide6.QtWidgets import QApplication, QWidget
from PySide6.QtGui import QColor, QPainter, QPen, QBrush, QFont
from PySide6.QtCore import QRectF, Qt

Number = Union[int, float]


# ============================================================
# 0) Helpers: CSS value parsing (px, %, rgba, shorthand...)
# ============================================================
def _to_str(v) -> str:
    return "" if v is None else str(v).strip()

def _parse_px(v, default: float = 0.0) -> float:
    if v is None:
        return default
    if isinstance(v, (int, float)):
        return float(v)
    s = _to_str(v).lower()
    if s.endswith("px"):
        s = s[:-2].strip()
    # allow "12.5"
    try:
        return float(s)
    except ValueError:
        return default

def _parse_percent(v, default: float = 0.0) -> float:
    # returns 0..1
    if v is None:
        return default
    if isinstance(v, (int, float)):
        # if they pass 0.5 treat as ratio
        x = float(v)
        return x if 0.0 <= x <= 1.0 else x / 100.0
    s = _to_str(v).lower()
    if s.endswith("%"):
        try:
            return float(s[:-1].strip()) / 100.0
        except ValueError:
            return default
    try:
        x = float(s)
        return x if 0.0 <= x <= 1.0 else x / 100.0
    except ValueError:
        return default

def _parse_color(v, default: str = "transparent") -> QColor:
    if v is None:
        return QColor(default)
    if isinstance(v, QColor):
        return QColor(v)
    s = _to_str(v)
    c = QColor(s)
    return c if c.isValid() else QColor(default)

def _parse_bool(v, default: bool = False) -> bool:
    if v is None:
        return default
    if isinstance(v, bool):
        return v
    s = _to_str(v).lower()
    if s in ("1", "true", "yes", "on"):
        return True
    if s in ("0", "false", "no", "off"):
        return False
    return default

def _parse_enum(v, allowed: set, default):
    s = _to_str(v).lower()
    return s if s in allowed else default

def _parse_spacing(v, default=(0, 0, 0, 0)) -> Tuple[float, float, float, float]:
    """
    CSS shorthand:
      - 1 value: all
      - 2 values: top/bottom, left/right
      - 3 values: top, left/right, bottom
      - 4 values: top, right, bottom, left
    Accepts number, "12px", "8 12", "8px 12px 4px 2px"
    Returns (top, right, bottom, left)
    """
    if v is None:
        return default
    if isinstance(v, (int, float)):
        x = float(v)
        return (x, x, x, x)

    s = _to_str(v)
    parts = s.replace(",", " ").split()
    nums = [_parse_px(p, None) for p in parts]
    nums = [x for x in nums if x is not None]

    if len(nums) == 1:
        a = nums[0]; return (a, a, a, a)
    if len(nums) == 2:
        a, b = nums; return (a, b, a, b)
    if len(nums) == 3:
        a, b, c = nums; return (a, b, c, b)
    if len(nums) >= 4:
        a, b, c, d = nums[:4]; return (a, b, c, d)
    return default

def _parse_border(v: Any) -> Tuple[float, str, QColor]:
    """
    Supports:
      - "2px solid white"
      - {"width":"2px","style":"solid","color":"white"}
    Returns: (width, style, color)
    """
    if v is None:
        return (0.0, "solid", QColor("transparent"))

    if isinstance(v, dict):
        w = _parse_px(v.get("width"), 0.0)
        style = _parse_enum(v.get("style"), {"none", "solid", "dashed", "dotted"}, "solid")
        c = _parse_color(v.get("color"), "transparent")
        return (w, style, c)

    s = _to_str(v)
    parts = s.split()
    w = 0.0
    style = "solid"
    c = QColor("white")
    for p in parts:
        pl = p.lower()
        if pl.endswith("px"):
            try:
                w = float(pl[:-2])
            except ValueError:
                pass
        elif pl in ("none", "solid", "dashed", "dotted"):
            style = pl
        else:
            cc = QColor(p)
            if cc.isValid():
                c = cc
    return (w, style, c)

def _parse_box_shadow(v: Any):
    """
    Subset: "offsetX offsetY blur spread color"
    Example: "0px 8px 24px 0px rgba(0,0,0,0.25)"
    We'll parse numbers + last token color (best-effort).
    Returns: (ox, oy, blur, spread, color, enabled)
    """
    if v is None:
        return (0.0, 0.0, 0.0, 0.0, QColor("transparent"), False)

    s = _to_str(v)
    parts = s.split()
    nums = []
    col = None
    for p in parts:
        if p.lower().endswith("px") or p.replace(".", "", 1).isdigit():
            nums.append(_parse_px(p, 0.0))
        else:
            # try color
            c = QColor(p)
            if c.isValid():
                col = c

    while len(nums) < 4:
        nums.append(0.0)
    ox, oy, blur, spread = nums[:4]
    color = col if col is not None else QColor(0, 0, 0, 80)
    return (ox, oy, blur, spread, color, True)


# ============================================================
# 1) Style model: CSS subset for QtMUI widgets
# ============================================================
@dataclass
class ComputedStyle:
    # size
    width: Optional[float] = None
    height: Optional[float] = None
    min_width: Optional[float] = None
    min_height: Optional[float] = None
    max_width: Optional[float] = None
    max_height: Optional[float] = None

    # box model
    margin: Tuple[float, float, float, float] = (0, 0, 0, 0)   # t r b l
    padding: Tuple[float, float, float, float] = (0, 0, 0, 0)  # t r b l

    # background
    background_color: QColor = field(default_factory=lambda: QColor("transparent"))

    # border
    border_width: float = 0.0
    border_style: str = "solid"     # none/solid/dashed/dotted
    border_color: QColor = field(default_factory=lambda: QColor("transparent"))
    border_radius: float = 0.0      # px only in this subset

    # outline (subset)
    outline_width: float = 0.0
    outline_color: QColor = field(default_factory=lambda: QColor("transparent"))

    # opacity / visibility
    opacity: float = 1.0
    visibility: str = "visible"     # visible/hidden
    overflow: str = "visible"       # visible/hidden

    # transform base from sx (static) + dynamic from variants
    translate_x: float = 0.0
    translate_y: float = 0.0
    rotate: float = 0.0             # degrees
    scale_x: float = 1.0
    scale_y: float = 1.0

    # shadow (subset)
    shadow_enabled: bool = False
    shadow: bool = False
    shadow_ox: float = 0.0
    shadow_oy: float = 0.0
    shadow_blur: float = 0.0
    shadow_spread: float = 0.0
    shadow_color: QColor = field(default_factory=lambda: QColor("transparent"))

    # typography (subset)
    color: QColor = field(default_factory=lambda: QColor("white"))
    font_size: float = 14.0
    font_weight: int = 400
    line_height: float = 1.2
    text_align: str = "left"        # left/center/right

    # misc (subset)
    cursor: str = "default"
    user_select: str = "auto"

    # store unknown props for later extension
    extra: Dict[str, Any] = field(default_factory=dict)


def parse_sx(sx: dict) -> ComputedStyle:
    st = ComputedStyle()

    # sizing
    if "width" in sx: st.width = _parse_px(sx.get("width"), None)
    if "height" in sx: st.height = _parse_px(sx.get("height"), None)
    if "minWidth" in sx or "min-width" in sx:
        st.min_width = _parse_px(sx.get("minWidth", sx.get("min-width")), None)
    if "minHeight" in sx or "min-height" in sx:
        st.min_height = _parse_px(sx.get("minHeight", sx.get("min-height")), None)
    if "maxWidth" in sx or "max-width" in sx:
        st.max_width = _parse_px(sx.get("maxWidth", sx.get("max-width")), None)
    if "maxHeight" in sx or "max-height" in sx:
        st.max_height = _parse_px(sx.get("maxHeight", sx.get("max-height")), None)

    # spacing
    if "margin" in sx: st.margin = _parse_spacing(sx.get("margin"), st.margin)
    if "padding" in sx: st.padding = _parse_spacing(sx.get("padding"), st.padding)

    # background
    if "background" in sx:
        # allow "background" as a color in this subset
        st.background_color = _parse_color(sx.get("background"), "transparent")
    if "background-color" in sx:
        st.background_color = _parse_color(sx.get("background-color"), "transparent")
    if "bgcolor" in sx:
        st.background_color = _parse_color(sx.get("bgcolor"), "transparent")

    # border
    if "border" in sx:
        bw, bstyle, bc = _parse_border(sx.get("border"))
        st.border_width, st.border_style, st.border_color = bw, bstyle, bc
    if "border-width" in sx: st.border_width = _parse_px(sx.get("border-width"), st.border_width)
    if "border-style" in sx: st.border_style = _parse_enum(sx.get("border-style"), {"none","solid","dashed","dotted"}, st.border_style)
    if "border-color" in sx: st.border_color = _parse_color(sx.get("border-color"), "transparent")

    if "border-radius" in sx: st.border_radius = _parse_px(sx.get("border-radius"), st.border_radius)
    if "borderRadius" in sx: st.border_radius = _parse_px(sx.get("borderRadius"), st.border_radius)
    
    # outline
    if "outline" in sx:
        # accept "2px solid red" as outline too
        ow, _, oc = _parse_border(sx.get("outline"))
        st.outline_width, st.outline_color = ow, oc
    if "outline-width" in sx: st.outline_width = _parse_px(sx.get("outline-width"), st.outline_width)
    if "outline-color" in sx: st.outline_color = _parse_color(sx.get("outline-color"), "transparent")

    # opacity / visibility / overflow
    if "opacity" in sx: st.opacity = float(sx.get("opacity"))
    if "visibility" in sx: st.visibility = _parse_enum(sx.get("visibility"), {"visible","hidden"}, st.visibility)
    if "overflow" in sx: st.overflow = _parse_enum(sx.get("overflow"), {"visible","hidden"}, st.overflow)

    # box-shadow
    if "box-shadow" in sx:
        ox, oy, blur, spread, col, enabled = _parse_box_shadow(sx.get("box-shadow"))
        st.shadow_enabled = enabled
        st.shadow_ox, st.shadow_oy, st.shadow_blur, st.shadow_spread, st.shadow_color = ox, oy, blur, spread, col
        st.shadow = sx.get("box-shadow")

    # transform base (static)
    if "translateX" in sx: st.translate_x = _parse_px(sx.get("translateX"), st.translate_x)
    if "translateY" in sx: st.translate_y = _parse_px(sx.get("translateY"), st.translate_y)
    if "rotate" in sx: st.rotate = float(sx.get("rotate"))
    if "scaleX" in sx: st.scale_x = float(sx.get("scaleX"))
    if "scaleY" in sx: st.scale_y = float(sx.get("scaleY"))
    if "scale" in sx:
        sc = float(sx.get("scale"))
        st.scale_x = sc
        st.scale_y = sc

    # typography
    if "color" in sx: st.color = _parse_color(sx.get("color"), "white")
    if "fontSize" in sx or "font-size" in sx:
        st.font_size = _parse_px(sx.get("fontSize", sx.get("font-size")), st.font_size)
    if "fontWeight" in sx or "font-weight" in sx:
        try:
            st.font_weight = int(float(sx.get("fontWeight", sx.get("font-weight"))))
        except Exception:
            pass
    if "lineHeight" in sx or "line-height" in sx:
        try:
            st.line_height = float(sx.get("lineHeight", sx.get("line-height")))
        except Exception:
            pass
    if "textAlign" in sx or "text-align" in sx:
        st.text_align = _parse_enum(sx.get("textAlign", sx.get("text-align")), {"left","center","right"}, st.text_align)

    # misc
    if "cursor" in sx: st.cursor = _to_str(sx.get("cursor"))
    if "userSelect" in sx or "user-select" in sx:
        st.user_select = _parse_enum(sx.get("userSelect", sx.get("user-select")), {"auto","none"}, st.user_select)

    # unknown props keep
    known = {
        "width","height","minWidth","min-width","minHeight","min-height","maxWidth","max-width","maxHeight","max-height",
        "margin","padding","background","background-color","bgcolor",
        "border","border-width","border-style","border-color","border-radius","borderRadius",
        "outline","outline-width","outline-color",
        "opacity","visibility","overflow",
        "box-shadow",
        "translateX","translateY","rotate","scale","scaleX","scaleY",
        "color","fontSize","font-size","fontWeight","font-weight","lineHeight","line-height","textAlign","text-align",
        "cursor","userSelect","user-select"
    }
    for k, v in sx.items():
        if k not in known:
            st.extra[k] = v

    return st


# ============================================================
# 2) Variants compiler (keyframes + cubic-bezier easing)
# ============================================================
@dataclass
class KeyframeTrack:
    times: List[float]
    values: List[Number]

def _lerp(a: float, b: float, t: float) -> float:
    return a + (b - a) * t

def _sample_track(track: KeyframeTrack, p: float) -> float:
    times = track.times
    vals = track.values
    for i in range(len(times) - 1):
        t0, t1 = times[i], times[i + 1]
        if p <= t1 or i == len(times) - 2:
            v0, v1 = float(vals[i]), float(vals[i + 1])
            if t1 == t0:
                return v1
            local = (p - t0) / (t1 - t0)
            local = max(0.0, min(1.0, local))
            return _lerp(v0, v1, local)
    return float(vals[-1])

def _cubic_bezier_ease(x1: float, y1: float, x2: float, y2: float, x: float) -> float:
    x = max(0.0, min(1.0, x))

    def bx(t: float) -> float:
        u = 1.0 - t
        return 3*u*u*t*x1 + 3*u*t*t*x2 + t*t*t

    def by(t: float) -> float:
        u = 1.0 - t
        return 3*u*u*t*y1 + 3*u*t*t*y2 + t*t*t

    def dbx(t: float) -> float:
        u = 1.0 - t
        return 3*u*u*x1 + 6*u*t*(x2 - x1) + 3*t*t*(1.0 - x2)

    t = x
    for _ in range(8):
        f = bx(t) - x
        d = dbx(t)
        if abs(d) < 1e-6:
            break
        t_next = t - f / d
        if t_next < 0.0 or t_next > 1.0:
            break
        t = t_next

    lo, hi = 0.0, 1.0
    for _ in range(12):
        mid = (lo + hi) * 0.5
        if bx(mid) < x:
            lo = mid
        else:
            hi = mid
    t = (lo + hi) * 0.5

    return max(0.0, min(1.0, by(t)))


class Timeline:
    def __init__(self, duration_ms: int, loop_count: int, easing: Union[QEasingCurve, List[float]]):
        self.duration_ms = duration_ms
        self.loop_count = loop_count
        self.easing = easing
        self.tracks: Dict[str, KeyframeTrack] = {}

    def add_track(self, name: str, times: List[float], values: List[Number]):
        if len(times) != len(values):
            raise ValueError(f"Track '{name}': times({len(times)}) != values({len(values)})")
        if len(times) < 2:
            raise ValueError(f"Track '{name}': need >=2 keyframes")
        if abs(times[0] - 0.0) > 1e-9 or abs(times[-1] - 1.0) > 1e-9:
            raise ValueError(f"Track '{name}': times must start at 0 and end at 1")
        for i in range(1, len(times)):
            if times[i] < times[i - 1]:
                raise ValueError(f"Track '{name}': times must be non-decreasing")
        self.tracks[name] = KeyframeTrack(times=times, values=values)

    def _ease(self, p: float) -> float:
        p = max(0.0, min(1.0, p))
        if isinstance(self.easing, QEasingCurve):
            return self.easing.valueForProgress(p)
        if isinstance(self.easing, list) and len(self.easing) == 4:
            x1, y1, x2, y2 = map(float, self.easing)
            return _cubic_bezier_ease(x1, y1, x2, y2, p)
        return p

    def sample(self, progress_0_1: float) -> Dict[str, float]:
        p = self._ease(float(progress_0_1))
        return {k: float(_sample_track(t, p)) for k, t in self.tracks.items()}


def compile_variant_timeline(
    variants: dict,
    variant_key: str,
    apply_fn: Callable[[Dict[str, float]], None],
    parent: Optional[QObject] = None,
) -> QVariantAnimation:
    node = variants.get(variant_key, {}) or {}
    trans = node.get("transition", {}) or {}

    dur = trans.get("duration", 1.0)
    # duration: float <= 20 => seconds; else treat as ms
    if isinstance(dur, (int, float)):
        duration_ms = int(round(float(dur) * 1000)) if float(dur) <= 20 else int(round(float(dur)))
    else:
        duration_ms = 1000

    count = trans.get("count", 1)
    loop = -1 if str(count).lower() in ("infinite", "-1") else int(count)

    ease = trans.get("ease", "easeInOut")
    if isinstance(ease, list) and len(ease) == 4 and all(isinstance(x, (int, float)) for x in ease):
        easing: Union[QEasingCurve, List[float]] = [float(x) for x in ease]
    else:
        s = str(ease).lower()
        easing = QEasingCurve(QEasingCurve.Linear) if s == "linear" else QEasingCurve(QEasingCurve.InOutQuad)

    times = trans.get("times", None)
    if isinstance(times, list) and len(times) >= 2 and all(isinstance(x, (int, float)) for x in times):
        times = [float(x) for x in times]
        if abs(times[0] - 0.0) > 1e-9 or abs(times[-1] - 1.0) > 1e-9:
            times = None
    else:
        times = None

    timeline = Timeline(duration_ms=duration_ms, loop_count=loop, easing=easing)

    for prop, vals in node.items():
        if prop == "transition":
            continue
        if not isinstance(vals, list) or len(vals) < 2:
            continue

        if times is not None and len(times) == len(vals):
            local_times = times
        else:
            n = len(vals)
            local_times = [i / (n - 1) for i in range(n)]

        # ensure endpoints
        if local_times[0] != 0.0:
            local_times = [0.0] + local_times
            vals = [vals[0]] + vals
        if local_times[-1] != 1.0:
            local_times = local_times + [1.0]
            vals = vals + [vals[-1]]

        timeline.add_track(prop, local_times, vals)

    anim = QVariantAnimation(parent=parent)
    anim.setStartValue(0.0)
    anim.setEndValue(1.0)
    anim.setDuration(timeline.duration_ms)
    anim.setLoopCount(timeline.loop_count)

    def on_changed(v):
        apply_fn(timeline.sample(float(v)))

    anim.valueChanged.connect(on_changed)
    return anim



from PySide6.QtWidgets import QFrame, QVBoxLayout, QApplication
from PySide6.QtCore import QSize

from ..alert.alert import Alert
from ..button import Button
from ..py_tool_button.py_tool_button import Iconify
from ..snackbar import Snackbar
from ..box import Box
from ..stack import QStack
from ..typography import Typography

class IconBox(QFrame):
    def __init__(self, context=None, path:str=None):
        super().__init__()

        self._context = context

        text = str(path.split("/")[-1].split('.')[0][:7] + '..').capitalize()

        self.setLayout(QVBoxLayout())
        self.layout().addWidget(
            Box(
                children=[
                    Button(
                        # variant="text",
                        size="large",
                        variant="soft",
                        startIcon=Iconify(icon=path, size=QSize(32, 32), fillColor="#919EAB"),
                        # startIcon=Iconify(icon=path, pixmap=True,width=20, height=20, fillColor="#919EAB"),
                        onClick=lambda icon_p = path: self._copy_to_clipboard(icon_p),
                        tooltip=path
                    ),
                    Typography(text=f"{text} ", align="center", variant="caption", color="#919EAB")
                ]
            )
        )

    def _copy_to_clipboard(self, icon_p):
        # Tạo đối tượng clipboard
        clipboard = QApplication.clipboard()
        clipboard.setText(icon_p)
        self._create_snackbar(f"Đã sao chép vào clipboard: {icon_p}")

    def _create_snackbar(self, message):
        snackbar = Snackbar(
            self._context, 
            message=message, 
            duration=3000, 
            position="top", 
            spacing=6,
            child=Alert(
                text="Đã sao chép vào clipboard"
            ),
        )
        snackbar.showSnackbar(self._context, spacing=6)


import asyncio
import uuid
from typing import Optional, Callable, Any, Dict, Union

from PySide6.QtWidgets import QLineEdit, QHBoxLayout, QGroupBox
from PySide6.QtCore import Signal, Qt, QTimer
from qtmui.material.styles import useTheme
from qtmui.material.styles.create_theme.components.get_qss_styles import get_qss_style

from qtmui.hooks import State

from .py_input import MuiInput

class MuiStandardInput(MuiInput):

    def __init__(
        self,
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
    
    # async def __set_stylesheet(self):
    def _set_stylesheet(self):
        super()._set_stylesheet()
        if not hasattr(self, "theme"):
            self.theme = useTheme()

        MuiStandardInput = self.theme.components["MuiStandardInput"].get("styles")
        MuiStandardInput_root = MuiStandardInput.get("root")
        MuiStandardInput_root_qss = get_qss_style(MuiStandardInput_root)
        MuiStandardInput_root_slot_hovered_qss = get_qss_style(MuiStandardInput_root["slots"]["hover"])
        MuiStandardInput_root_slot_focused_qss = get_qss_style(MuiStandardInput_root["slots"]["focus"])
        
        MuiStandardInput_title = MuiStandardInput.get("title")
        MuiStandardInput_title_qss = get_qss_style(MuiStandardInput_title)

        MuiStandardInput_inputField_root = MuiStandardInput.get("inputField")["root"]
        MuiStandardInput_inputField_root_qss = get_qss_style(MuiStandardInput_inputField_root)

        MuiStandardInput_root_prop_hasStartAdornment_qss = ""
        # MuiStandardInput_root_prop_hasStartAdornment_qss = get_qss_style(MuiStandardInput_inputField_root["props"]["hasStartAdornment"])

        self.setProperty("variant", "standard")

        self.setStyleSheet(
            self.styleSheet() + 
            f"""
                #{self.objectName()}[variant=standard] {{
                    {MuiStandardInput_root_qss}
                }}
                #{self.objectName()}[variant=standard]:title {{
                    {MuiStandardInput_title_qss}
                }}
                #{self.objectName()}[variant=standard]  QLineEdit, QPlainTextEdit, QTextEdit {{
                    {MuiStandardInput_inputField_root_qss}
                }}
                #{self.objectName()}[standardHasStartAdornment=true]  QLineEdit, QPlainTextEdit, QTextEdit {{
                    {MuiStandardInput_root_prop_hasStartAdornment_qss}
                }}
                #{self.objectName()}[hovered=true] {{
                    {MuiStandardInput_root_slot_hovered_qss}
                }}
                #{self.objectName()}[focused=true] {{
                    {MuiStandardInput_root_slot_focused_qss}
                }}
            """
        )

    def _set_prop_has_start_adornment(self, state:bool):
        self.setProperty("standardHasStartAdornment", state)
        self._set_stylesheet()




import sys
from PySide6.QtCore import QObject, Property, Signal

class TreeModel(QObject):
    dataChanged = Signal()

    def __init__(self, data=[]):
        super().__init__()
        self._data = data

    def get_data(self):
        return self._data

    def set_data(self, value):
        if self._data != value:
            self._data = value
            self.dataChanged.emit()

    data = Property(str, get_data, set_data, notify=dataChanged)



# Copyright (C) 2022 The Pyact Company Ltd.
# SPDX-License-Identifier: LicenseRef-Pyact-Commercial OR BSD-3-Clause
from __future__ import annotations

from typing import Optional, Union
from PySide6.QtCore import QPointF, Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout
from PySide6.QtCharts import (
    QChart,
    QChartView,
    QSplineSeries,
    QBarCategoryAxis,
    QValueAxis
)
from PySide6.QtGui import QPen, QPainter, QColor

from qtmui.material.styles import useTheme


class ChartSLine(QWidget):
    def __init__(
        self,
        dir: str = "ltr",
        type: str = "line",
        series: object = None,
        width: Optional[Union[str, int]] = None,
        height: Optional[Union[str, int]] = None,
        options: object = None,
        key: str = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._series_input = series or []
        self._options = options or {}
        self._width = width
        self._height = height

        self._init_chart()

        # Apply theme
        theme = useTheme()
        theme.state.valueChanged.connect(self._apply_theme)
        self._apply_theme()

    # -----------------------------------------------------------------------

    def _init_chart(self):
        if self._height:
            self.setFixedHeight(self._height)
        if self._width:
            self.setFixedWidth(self._width)

        # -----------------------------
        # Create spline series from data
        # -----------------------------
        self._spline_series_list = []

        for item in self._series_input:
            s = QSplineSeries()
            s.setName(item.get("name", ""))

            for idx, y in enumerate(item.get("data", [])):
                x = float(idx)
                y = float(y)
                s.append(QPointF(x, y))

            self._spline_series_list.append(s)

        # -----------------------------
        # Create Chart
        # -----------------------------
        self.chart = QChart()
        self.chart.legend().setVisible(True)
        self.chart.legend().setAlignment(Qt.AlignTop)

        for s in self._spline_series_list:
            self.chart.addSeries(s)

        # Default numeric axes
        axis_x = QValueAxis()
        axis_y = QValueAxis()

        self.chart.addAxis(axis_x, Qt.AlignBottom)
        self.chart.addAxis(axis_y, Qt.AlignLeft)

        # Attach all series to axes
        for s in self._spline_series_list:
            s.attachAxis(axis_x)
            s.attachAxis(axis_y)

        # -----------------------------
        # CATEGORY AXIS (if provided)
        # -----------------------------
        categories = None

        if self._options.get("xaxis") and "categories" in self._options["xaxis"]:
            categories = self._options["xaxis"]["categories"]

        if categories:
            cat_axis = QBarCategoryAxis()
            cat_axis.append(categories)
            self.chart.removeAxis(axis_x)
            self.chart.addAxis(cat_axis, Qt.AlignBottom)

            for s in self._spline_series_list:
                s.attachAxis(cat_axis)

            # Fix Y range
            axis_y.setRange(0, max(max(item["data"]) for item in self._series_input))

        # -----------------------------
        # Chart View
        # -----------------------------
        self.chart_view = QChartView(self.chart)
        self.chart_view.setRenderHint(QPainter.Antialiasing)
        self.layout().addWidget(self.chart_view)

    # -----------------------------------------------------------------------

    def _apply_theme(self):
        theme = useTheme()

        text_color = QColor(theme.palette.text.secondary)
        primary_color = QColor(theme.palette.primary.main)

        # Apply to title
        self.chart.setTitleBrush(text_color)

        # Axis color
        for axis in self.chart.axes():
            axis.setLabelsBrush(text_color)
            axis.setTitleBrush(text_color)
            axis.setLinePen(QPen(text_color))

        # Legend
        legend = self.chart.legend()
        legend.setLabelBrush(text_color)

        # Apply colors to spline series
        for idx, s in enumerate(self._spline_series_list):
            pen = QPen(primary_color if idx == 0 else QColor(theme.palette.info.main))
            pen.setWidth(3)
            s.setPen(pen)



from typing import Callable, List, Optional, Union, Dict, Any

from qtmui.hooks import State

from PySide6.QtCore import (
    Qt,
    QSize,
    Signal,
    QPoint,
    QPointF,
    QRectF,
    QEasingCurve,
    QPropertyAnimation,
    QSequentialAnimationGroup,
    Slot,
    Property,
)

from PySide6.QtWidgets import QCheckBox, QWidget
from PySide6.QtGui import QColor, QBrush, QPaintEvent, QPen, QPainter, QEnterEvent

from qtmui.material.styles import useTheme

from ..widget_base import PyWidgetBase
from ..utils.validate_params import _validate_param


class Switch(QCheckBox):
    _transparent_pen = QPen(Qt.transparent)

    VALID_COLORS = ['default', 'primary', 'secondary', 'error', 'info', 'success', 'warning']
    VALID_SIZES = ['small', 'medium']
    VALID_EDGES = ['start', 'end', False]
    SIZE_MAP = {
        "small": {
            "size": QSize(46, int(46 * 45 / 58)),
            "bar_ratio": 0.48,
            "handle_ratio": 0.18,
        },
        "medium": {
            "size": QSize(58, 45),
            "bar_ratio": 0.48,
            "handle_ratio": 0.20,
        },
    }

    def __init__(
        self,
        parent=None,
        label: Optional[Union[str, State, Callable]] = None,
        disabled: Union[State, bool]=False,
        defaultChecked: Union[State, bool]=False,
        checked: Union[State, bool]=False,
        onChange: Optional[Union[State, Callable, None]]=None,
        highlight: Union[State, bool]=False,
        left: Union[State, bool]=True,
        color: Optional[Union[State, str, None]]="primary",
        size: Optional[Union[State, str, None]]="medium",
        onClick: Optional[Union[State, Callable, None]]=None,
        checkedIcon: Optional[Union[State, QWidget, None]]=None,
        icon: Optional[Union[State, QWidget, None]]=None,
        classes: Optional[Union[State, Dict, None]]=None,
        sx: Optional[Union[State, List, Dict, Callable, None]]=None,
        disableRipple: Union[State, bool]=False,
        edge: Optional[Union[State, str, bool, None]]=False,
        id: Optional[Union[State, str, None]]=None,
        inputProps: Optional[Union[State, Dict, None]]=None,
        inputRef: Optional[Union[State, Any, None]]=None,
        required: Union[State, bool]=False,
        slotProps: Optional[Union[State, Dict, None]]=None,
        slots: Optional[Union[State, Dict, None]]=None,
        value: Optional[Union[State, Any, None]]=None,
        **kwargs
        
    ):
        super().__init__(parent)
        

        self._key = None
        self._color = color
        self._size = size
        self._value = None

        # Set properties with validation
        self._set_label(label)
        self._set_disabled(disabled)
        self._set_defaultChecked(defaultChecked)
        self._set_checked(checked)
        self._set_onChange(onChange)
        self._set_highlight(highlight)
        self._set_left(left)
        self._set_color(color)
        self._set_size(size)
        self._set_onClick(onClick)
        self._set_checkedIcon(checkedIcon)
        self._set_icon(icon)
        self._set_classes(classes)
        self._set_sx(sx)
        self._set_disableRipple(disableRipple)
        self._set_edge(edge)
        self._set_id(id)
        self._set_inputProps(inputProps)
        self._set_inputRef(inputRef)
        self._set_required(required)
        self._set_slotProps(slotProps)
        self._set_slots(slots)
        self._set_value(value)

        self._init_ui()

    # Setter and Getter methods
    @_validate_param(file_path="qtmui.material.switch", param_name="label", supported_signatures=Union[State, str, type(None)])
    def _set_label(self, value):
        self._label = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_label(self):
        return self._label.value if isinstance(self._label, State) else self._label

    @_validate_param(file_path="qtmui.material.switch", param_name="disabled", supported_signatures=Union[State, bool])
    def _set_disabled(self, value):
        self._disabled = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_disabled(self):
        return self._disabled.value if isinstance(self._disabled, State) else self._disabled

    @_validate_param(file_path="qtmui.material.switch", param_name="defaultChecked", supported_signatures=Union[State, bool])
    def _set_defaultChecked(self, value):
        self._defaultChecked = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_defaultChecked(self):
        return self._defaultChecked.value if isinstance(self._defaultChecked, State) else self._defaultChecked

    # @_validate_param(file_path="qtmui.material.switch", param_name="checked", supported_signatures=Union[State, bool])
    def _set_checked(self, value):
        self._checked = value

    def _get_checked(self):
        return self._checked.value if isinstance(self._checked, State) else self._checked

    @_validate_param(file_path="qtmui.material.switch", param_name="onChange", supported_signatures=Union[State, Callable, type(None)])
    def _set_onChange(self, value):
        self._onChange = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_onChange(self):
        return self._onChange.value if isinstance(self._onChange, State) else self._onChange

    @_validate_param(file_path="qtmui.material.switch", param_name="highlight", supported_signatures=Union[State, bool])
    def _set_highlight(self, value):
        self._highlight = value
        if isinstance(value, State):
            value.valueChanged.connect(self._set_stylesheet)

    def _get_highlight(self):
        return self._highlight.value if isinstance(self._highlight, State) else self._highlight

    @_validate_param(file_path="qtmui.material.switch", param_name="left", supported_signatures=Union[State, bool])
    def _set_left(self, value):
        self._left = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_left(self):
        return self._left.value if isinstance(self._left, State) else self._left

    @_validate_param(file_path="qtmui.material.switch", param_name="color", supported_signatures=Union[State, str, type(None)], valid_values=VALID_COLORS)
    def _set_color(self, value):
        self._color = value
        if isinstance(value, State):
            value.valueChanged.connect(self._set_stylesheet)

    def _get_color(self):
        color = self._color.value if isinstance(self._color, State) else self._color
        return color if color in self.VALID_COLORS or isinstance(color, str) else 'primary'

    @_validate_param(file_path="qtmui.material.switch", param_name="size", supported_signatures=Union[State, str, type(None)], valid_values=VALID_SIZES)
    def _set_size(self, value):
        self._size = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_size(self):
        size = self._size.value if isinstance(self._size, State) else self._size
        return size if size in self.VALID_SIZES else 'small'

    @_validate_param(file_path="qtmui.material.switch", param_name="onClick", supported_signatures=Union[State, Callable, type(None)])
    def _set_onClick(self, value):
        self._onClick = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_onClick(self):
        return self._onClick.value if isinstance(self._onClick, State) else self._onClick

    @_validate_param(file_path="qtmui.material.switch", param_name="checkedIcon", supported_signatures=Union[State, QWidget, type(None)])
    def _set_checkedIcon(self, value):
        self._checkedIcon = value
        if isinstance(value, QWidget):
            self._widget_references.append(value)
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_checkedIcon(self):
        return self._checkedIcon.value if isinstance(self._checkedIcon, State) else self._checkedIcon

    @_validate_param(file_path="qtmui.material.switch", param_name="icon", supported_signatures=Union[State, QWidget, type(None)])
    def _set_icon(self, value):
        self._icon = value
        if isinstance(value, QWidget):
            self._widget_references.append(value)
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_icon(self):
        return self._icon.value if isinstance(self._icon, State) else self._icon

    @_validate_param(file_path="qtmui.material.switch", param_name="classes", supported_signatures=Union[State, Dict, type(None)])
    def _set_classes(self, value):
        self._classes = value
        if isinstance(value, State):
            value.valueChanged.connect(self._set_stylesheet)

    def _get_classes(self):
        return self._classes.value if isinstance(self._classes, State) else self._classes

    @_validate_param(file_path="qtmui.material.switch", param_name="sx", supported_signatures=Union[State, List, Dict, Callable, type(None)])
    def _set_sx(self, value):
        self._sx = value
        if isinstance(value, State):
            value.valueChanged.connect(self._set_stylesheet)

    def _get_sx(self):
        return self._sx.value if isinstance(self._sx, State) else self._sx

    @_validate_param(file_path="qtmui.material.switch", param_name="disableRipple", supported_signatures=Union[State, bool])
    def _set_disableRipple(self, value):
        self._disableRipple = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_disableRipple(self):
        return self._disableRipple.value if isinstance(self._disableRipple, State) else self._disableRipple

    @_validate_param(file_path="qtmui.material.switch", param_name="edge", supported_signatures=Union[State, str, bool, type(None)], valid_values=VALID_EDGES)
    def _set_edge(self, value):
        self._edge = value
        if isinstance(value, State):
            value.valueChanged.connect(self._set_stylesheet)

    def _get_edge(self):
        edge = self._edge.value if isinstance(self._edge, State) else self._edge
        return edge if edge in self.VALID_EDGES else False

    @_validate_param(file_path="qtmui.material.switch", param_name="id", supported_signatures=Union[State, str, type(None)])
    def _set_id(self, value):
        self._id = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_id(self):
        return self._id.value if isinstance(self._id, State) else self._id

    @_validate_param(file_path="qtmui.material.switch", param_name="inputProps", supported_signatures=Union[State, Dict, type(None)])
    def _set_inputProps(self, value):
        self._inputProps = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_inputProps(self):
        return self._inputProps.value if isinstance(self._inputProps, State) else self._inputProps

    # @_validate_param(file_path="qtmui.material.switch", param_name="inputRef", supported_signatures=Union[State, type(None)])
    def _set_inputRef(self, value):
        self._inputRef = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_inputRef(self):
        return self._inputRef.value if isinstance(self._inputRef, State) else self._inputRef

    @_validate_param(file_path="qtmui.material.switch", param_name="required", supported_signatures=Union[State, bool])
    def _set_required(self, value):
        self._required = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_required(self):
        return self._required.value if isinstance(self._required, State) else self._required

    @_validate_param(file_path="qtmui.material.switch", param_name="slotProps", supported_signatures=Union[State, Dict, type(None)])
    def _set_slotProps(self, value):
        self._slotProps = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_slotProps(self):
        return self._slotProps.value if isinstance(self._slotProps, State) else self._slotProps

    @_validate_param(file_path="qtmui.material.switch", param_name="slots", supported_signatures=Union[State, Dict, type(None)])
    def _set_slots(self, value):
        self._slots = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_slots(self):
        return self._slots.value if isinstance(self._slots, State) else self._slots

    @_validate_param(file_path="qtmui.material.switch", param_name="value", supported_signatures=Union[State, type(None)])
    def _set_value(self, value):
        self._value = value
        if isinstance(value, State):
            value.valueChanged.connect(self._init_ui)

    def _get_value(self):
        return self._value.value if isinstance(self._value, State) else self._value
    
    def _init_ui(self):
        
        self._cfg = self.SIZE_MAP[self._size]

        # STATE
        self._handle_position = 0.0
        self._pulse_radius = 0.0
        self._hover_opacity = 0.0

        # ANIMATIONS
        self.animation = QPropertyAnimation(self, b"handle_position", self)
        self.animation.setDuration(150)
        self.animation.setEasingCurve(QEasingCurve.InOutCubic)

        self.pulse_anim = QPropertyAnimation(self, b"pulse_radius", self)
        self.pulse_anim.setDuration(300)

        self.hover_anim = QPropertyAnimation(self, b"hover_opacity", self)
        self.hover_anim.setDuration(160)
        self.hover_anim.setEasingCurve(QEasingCurve.OutCubic)

        self.anim_group = QSequentialAnimationGroup(self)
        self.anim_group.addAnimation(self.animation)
        self.anim_group.addAnimation(self.pulse_anim)

        self.stateChanged.connect(self._setup_animation)
        
        self.setCursor(Qt.PointingHandCursor)
        self.setContentsMargins(4, 0, 4, 0)
        self.setMouseTracking(True)
        
        self._set_stylesheet()
        
        # tạo con trỏ chuột hình bàn tay khi hover chuột qua Button
        if self._disabled == True:
            self.setEnabled(False)
        else:
            self.setCursor(Qt.PointingHandCursor)

        if self._defaultChecked:
            self.setChecked(True)

        if self._checked:
            self.setChecked(True)

        # gán sau khi đã setChecked ban đầu nếu có
        if self._get_onChange():
            self.stateChanged.connect(self._get_onChange())
            
        

    # --------------------------------------------------
    def _set_stylesheet(self):
        self.theme = useTheme()

        theme_components = self.theme.components

        # print('self._color___________________', self._color)
        PySwitchRootColor = theme_components[f"PySwitch"].get("styles").get("root").get(self._color)
        if theme_components[f"PySwitch"].get("styleOverrides"):
            PySwitchOverrideRootColor = theme_components[f"PySwitch"].get("styleOverrides").get("root").get(self._color)

        barColor = PySwitchRootColor["barColor"]
        handleColor = PySwitchRootColor["handleColor"]
        checkedColor = PySwitchRootColor["checkedColor"]
        pulseCheckedColor = PySwitchRootColor["pulseCheckedColor"]
        pulseUncheckedColor = PySwitchRootColor["pulseUncheckedColor"]

        # if self._disabled:
        #     barColor = self.theme.palette.grey._400 if self.theme.palette.mode == "light" else self.theme.palette.grey._600
            
        self._bar_brush = QBrush(barColor)
        self._bar_checked_brush = QBrush(checkedColor)

        self._handle_brush = QBrush(QColor(handleColor))
        self._handle_checked_brush = QBrush(QColor(handleColor))

        # self._pulse_unchecked_brush = QBrush(QColor(self.theme.palette.grey._100))
        self._pulse_unchecked_brush = QBrush(QColor(self.theme.palette.grey._500).lighter(120))
        self._pulse_checked_brush = QBrush(QColor(pulseCheckedColor))
        
        self._hover_unchecked_brush = QBrush(QColor(self.theme.palette.grey._500).lighter(120))
        self._hover_checked_brush = QBrush(QColor(pulseCheckedColor).lighter(120))


    # --------------------------------------------------
    def sizeHint(self):
        return self._cfg["size"] if hasattr(self, "_cfg") else QSize(58, 45)


    def hitButton(self, pos: QPoint):
        return self.contentsRect().contains(pos)

    @Slot(int)
    def _setup_animation(self, value):

        self.anim_group.stop()

        contRect = self.contentsRect()
        handle_radius = self._cfg["handle_ratio"] * contRect.height()

        self.animation.setEndValue(1.0 if value else 0.0)

        # if self._init_done:
        if hasattr(self, "_hovered"):
            self.pulse_anim.setStartValue(handle_radius * 1.1)
            self.pulse_anim.setEndValue(min(contRect.height() / 2, handle_radius * 2.2))

        self.anim_group.start()

    # --------------------------------------------------
    def enterEvent(self, event: QEnterEvent):
        self._hovered = True
        self.hover_anim.stop()
        self.hover_anim.setEndValue(1.0)
        self.hover_anim.start()
        super().enterEvent(event)

    def leaveEvent(self, event):
        self.hover_anim.stop()
        self.hover_anim.setEndValue(0.0)
        self.hover_anim.start()
        super().leaveEvent(event)

    # --------------------------------------------------
    def paintEvent(self, e: QPaintEvent):
        if not hasattr(self, "_bar_brush"):
            return
        
        contRect = self.contentsRect()

        bar_height = self._cfg["bar_ratio"] * contRect.height()
        bar_radius = bar_height / 2
        handle_radius = self._cfg["handle_ratio"] * contRect.height()

        # 👉 SAFE RADIUS: bảo đảm hover không tràn
        hover_radius = handle_radius * 2.2

        p = QPainter(self)
        p.setRenderHint(QPainter.Antialiasing)
        p.setPen(self._transparent_pen)

        # ---- BAR (TRACK) ----
        bar_left = contRect.x() + handle_radius
        bar_width = contRect.width() - 2 * handle_radius

        barRect = QRectF(bar_left, 0, bar_width, bar_height)
        barRect.moveCenter(contRect.center())

        # ---- HANDLE TRACK (CHUẨN MUI) ----
        left_x = barRect.left() + bar_radius
        right_x = barRect.right() - bar_radius

        xPos = left_x + (right_x - left_x) * self._handle_position
        yPos = barRect.center().y()

        # ---- HOVER SOFT ----
        if self._hover_opacity > 0:
            if self.isChecked():
                p.setBrush(self._hover_checked_brush)
            else:
                p.setBrush(self._hover_unchecked_brush)
                
            p.setOpacity(self._hover_opacity * 0.35)
            p.drawEllipse(QPointF(xPos, yPos), hover_radius, hover_radius)
            p.setOpacity(1.0)

        # ---- PULSE ----
        if self.pulse_anim.state() == QPropertyAnimation.Running and hasattr(self, "_hovered"):
            p.setBrush(
                self._pulse_checked_brush
                if self.isChecked()
                else self._pulse_unchecked_brush
            )
            p.drawEllipse(QPointF(xPos, yPos), self._pulse_radius, self._pulse_radius)

        # ---- BAR ----
        p.setBrush(self._bar_checked_brush if self.isChecked() else self._bar_brush)
        p.drawRoundedRect(barRect, bar_radius, bar_radius)

        # ---- HANDLE ----
        p.setBrush(self._handle_checked_brush if self.isChecked() else self._handle_brush)
        p.drawEllipse(QPointF(xPos, yPos), handle_radius, handle_radius)

        p.end()

    # --------------------------------------------------
    @Property(float)
    def handle_position(self):
        return self._handle_position

    @handle_position.setter
    def handle_position(self, v):
        self._handle_position = v
        self.update()

    @Property(float)
    def pulse_radius(self):
        return self._pulse_radius

    @pulse_radius.setter
    def pulse_radius(self, v):
        self._pulse_radius = v
        self.update()

    @Property(float)
    def hover_opacity(self):
        return self._hover_opacity

    @hover_opacity.setter
    def hover_opacity(self, v):
        self._hover_opacity = v
        self.update()



from typing import Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from qtmui.material.styles.create_theme.theme_reducer import ThemeState

from .properties_name import *
from ....system.color_manipulator import alpha


def stepper(_theme) -> Dict:
    theme: ThemeState = _theme
    # theme = _theme
    return {
        'PyStepper': {
            'styles': {
                "root": {
                    backgroundColor: "transparent"
                },
            },
        },
    }



import uuid
from typing import Callable, Union, List, Optional, Literal
from PySide6.QtWidgets import QWidget, QVBoxLayout, QSizePolicy, QApplication
from PySide6.QtCore import QEvent, Qt, QObject
from qtmui.hooks import State
from qtmui.material.styles import useTheme
from ..utils.validate_params import _validate_param

class ClickAwayListener(QWidget):
    """
    A component that listens for click or touch events outside of its children and triggers a callback.

    The `ClickAwayListener` component wraps a widget or a list of widgets and detects click or touch
    events that occur outside of its boundaries, triggering the `onClickAway` callback. It supports
    all props of the Material-UI `ClickAwayListener` component.

    Parameters
    ----------
    children : State, QWidget, List[QWidget], or None, optional
        The wrapped element(s), which must be able to hold a reference (QWidget or list of QWidgets).
        Default is None. Can be a `State` object for dynamic updates.
    onClickAway : State, Callable, or None, optional
        Callback fired when a click or touch event is detected outside the children.
        Signature: function(event: QEvent) -> None
        Default is None. Can be a `State` object for dynamic updates.
    disableReactTree : State or bool, optional
        If True, only the DOM tree is considered, ignoring the React tree. In Qt, this limits event
        checking to the widget's direct children. Default is False.
        Can be a `State` object for dynamic updates.
    mouseEvent : State, str, or Literal[False], optional
        The mouse event to listen to. Supported values: "onClick", "onMouseDown", "onMouseUp",
        "onPointerDown", "onPointerUp", or False to disable. Default is "onClick".
        Can be a `State` object for dynamic updates.
    touchEvent : State, str, or Literal[False], optional
        The touch event to listen to. Supported values: "onTouchEnd", "onTouchStart", or False to
        disable. Default is "onTouchEnd".
        Can be a `State` object for dynamic updates.
    **kwargs
        Additional keyword arguments passed to the parent `QWidget` class.

    Attributes
    ----------
    VALID_MOUSE_EVENTS : list[Union[str, bool]]
        Valid values for the `mouseEvent` parameter: ["onClick", "onMouseDown", "onMouseUp",
        "onPointerDown", "onPointerUp", False].
    VALID_TOUCH_EVENTS : list[Union[str, bool]]
        Valid values for the `touchEvent` parameter: ["onTouchEnd", "onTouchStart", False].

    Notes
    -----
    - The `children` prop must be a `QWidget` or a list of `QWidget` instances that can hold a reference.
    - The `onClickAway` callback receives a `QEvent` object as its argument.
    - The `disableReactTree` prop limits event checking to the widget's direct children when True.

    Demos:
    - ClickAwayListener: https://qtmui.com/material-ui/qtmui-click-away-listener/

    API Reference:
    - ClickAwayListener API: https://qtmui.com/material-ui/api/click-away-listener/
    """

    VALID_MOUSE_EVENTS = ["onClick", "onMouseDown", "onMouseUp", "onPointerDown", "onPointerUp", False]
    VALID_TOUCH_EVENTS = ["onTouchEnd", "onTouchStart", False]

    def __init__(
        self,
        children: Optional[Union[State, QWidget, List[QWidget]]] = None,
        onClickAway: Optional[Union[State, Callable]] = None,
        disableReactTree: Union[State, bool] = False,
        mouseEvent: Union[State, str, Literal[False]] = "onClick",
        touchEvent: Union[State, str, Literal[False]] = "onTouchEnd",
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.setObjectName(str(uuid.uuid4()))

        # Set properties with validation
        self._set_children(children)
        self._set_onClickAway(onClickAway)
        self._set_disableReactTree(disableReactTree)
        self._set_mouseEvent(mouseEvent)
        self._set_touchEvent(touchEvent)

        self._init_ui()

    # Setter and Getter methods for all parameters
    @_validate_param(file_path="qtmui.material.click_away_listener", param_name="children", supported_signatures=Union[State, QWidget, List[QWidget], type(None)])
    def _set_children(self, value):
        if isinstance(value, list) and not all(isinstance(item, QWidget) for item in value):
            raise ValueError("children must be a QWidget or a list of QWidgets")
        if value is not None and not isinstance(value, (QWidget, list, State)):
            raise ValueError("children must be a QWidget, list of QWidgets, or State")
        self._children = value

    def _get_children(self):
        return self._children.value if isinstance(self._children, State) else self._children

    @_validate_param(file_path="qtmui.material.click_away_listener", param_name="onClickAway", supported_signatures=Union[State, Callable, type(None)])
    def _set_onClickAway(self, value):
        self._onClickAway = value

    def _get_onClickAway(self):
        return self._onClickAway.value if isinstance(self._onClickAway, State) else self._onClickAway

    @_validate_param(file_path="qtmui.material.click_away_listener", param_name="disableReactTree", supported_signatures=Union[State, bool])
    def _set_disableReactTree(self, value):
        self._disableReactTree = value

    def _get_disableReactTree(self):
        return self._disableReactTree.value if isinstance(self._disableReactTree, State) else self._disableReactTree

    @_validate_param(file_path="qtmui.material.click_away_listener", param_name="mouseEvent", supported_signatures=Union[State, str, Literal[False]], valid_values=VALID_MOUSE_EVENTS)
    def _set_mouseEvent(self, value):
        self._mouseEvent = value

    def _get_mouseEvent(self):
        return self._mouseEvent.value if isinstance(self._mouseEvent, State) else self._mouseEvent

    @_validate_param(file_path="qtmui.material.click_away_listener", param_name="touchEvent", supported_signatures=Union[State, str, Literal[False]], valid_values=VALID_TOUCH_EVENTS)
    def _set_touchEvent(self, value):
        self._touchEvent = value

    def _get_touchEvent(self):
        return self._touchEvent.value if isinstance(self._touchEvent, State) else self._touchEvent

    def _init_ui(self):
        self.setObjectName(str(uuid.uuid4()))
        # self.setStyleSheet('''#{}  {{ {} }};'''.format(self.objectName(), "background-color: red"))
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0,0,0,0)
        # self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)

        if isinstance(self._children, list):
            for widget in self._children:
                if isinstance(widget, QWidget):
                    self.layout().addWidget(widget)

        self.installEventFilter(QApplication.instance().mainWindow)

    def eventFilter(self, obj, event: QEvent):
        if event.type() == QEvent.MouseButtonPress:
            if event.button() == Qt.LeftButton and not self.underMouse():
                if self._onClickAway:
                    self._onClickAway()

        return super().eventFilter(obj, event)


from PySide6.QtWidgets import QApplication, QMainWindow, QTreeView, QFileSystemModel
from PySide6.QtCore import Qt

class TreeViewModel(QFileSystemModel):
    def __init__(self, 
                 path, # QFileSystemModel
                 **kwargs
                 ):
        super().__init__(**kwargs)

        # Thiết lập model tệp hệ thống
        self.setRootPath(path)  # Gán root cho model (ví dụ: thư mục dự án)




from PySide6.QtWidgets import (
    QApplication, QListView, QFrame, QHBoxLayout, QLabel, QCheckBox, QWidget, QVBoxLayout, QStyledItemDelegate, QComboBox
)
from PySide6.QtCore import Qt, QAbstractListModel, QModelIndex, QSize

from typing import List
import sys

# ------------------------------
# Custom Widget Item
# ------------------------------
class CustomMenuItem(QFrame):
    def __init__(self, children: List[QWidget]):
        super().__init__()
        layout = QHBoxLayout(self)
        layout.setContentsMargins(4, 2, 4, 2)
        for widget in children:
            layout.addWidget(widget)


# ------------------------------
# Data model
# ------------------------------
class MyListModel(QAbstractListModel):
    def __init__(self, data):
        super().__init__()
        self._data = data  # mỗi item là dict: {'selected': bool}

    def rowCount(self, parent=QModelIndex()):
        return len(self._data)

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.UserRole:
            return self._data[index.row()]
        return None

    def toggle_selected(self, index: QModelIndex):
        row = index.row()
        self._data[row]["selected"] = not self._data[row]["selected"]
        self.dataChanged.emit(index, index, [Qt.UserRole])


# ------------------------------
# Delegate để tạo widget custom
# ------------------------------
class MyItemDelegate(QStyledItemDelegate):
    def __init__(self, parent=None, model=None):
        super(MyItemDelegate, self).__init__(parent)
        self._model = model

    def createEditor(self, parent, option, index):
        item_data = index.model().data(index, Qt.UserRole)
        row = index.row()
        selected = item_data.get("selected", False)

        print(row)

        # Widget hiển thị
        editor = CustomMenuItem([
            QCheckBox(checked=selected),
            QLabel(f"Item {row}")
        ])
        editor.setParent(parent)
        return editor

    def paint(self, painter, option, index):
        super(MyItemDelegate, self).paint(painter, option, index)
        if not self.parent().indexWidget(index) and not index.parent().isValid():

            item_data = index.model().data(index, Qt.UserRole)

            index_row = index.row()
            index_col = index.column()

            selected = item_data.get("selected", False)

            editor = CustomMenuItem([
                QCheckBox(checked=selected),
                QLabel(f"Item {index_row}")
            ])
            self.parent().setIndexWidget(index, editor)

    def sizeHint(self, option, index):
        return QSize(200, 32)


class Select(QComboBox):
    def __init__():
        super().__init__()


# ------------------------------
# Main Window
# ------------------------------
class MainWindow(QWidget):
    def __init__(self):
        super().__init__()

        # Dữ liệu mẫu
        data = [{"selected": True} for _ in range(10)]

        # Model & View
        self.model = MyListModel(data)
        self.list_view = QListView()
        self.list_view.setModel(self.model)
        self.list_view.setItemDelegate(MyItemDelegate(parent=self.list_view))
        self.list_view.setEditTriggers(QListView.AllEditTriggers)  # luôn hiển thị widget custom

        # Layout
        layout = QVBoxLayout(self)
        layout.addWidget(self.list_view)
        layout.addWidget(self.list_view)

        # Cho phép click để toggle selected
        self.list_view.clicked.connect(self.on_item_clicked)

    def on_item_clicked(self, index: QModelIndex):
        self.model.toggle_selected(index)


# ------------------------------
# App
# ------------------------------
if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = MainWindow()
    window.resize(300, 400)
    window.show()
    sys.exit(app.exec())



from typing import Optional, Union, Callable, Dict, List, Any
import uuid
from PySide6.QtWidgets import QFrame, QHBoxLayout, QSizePolicy
from PySide6.QtCore import QSize, Qt
from PySide6.QtGui import QCursor
from ..system.color_manipulator import lighten, rgbToHex
from ..py_tool_button import PyToolButton
from ..py_iconify.py_iconify import PyIconify
from ..widget_base import PyWidgetBase
from qtmui.material.styles import useTheme
from qtmui.material.styles.create_theme.components.get_qss_styles import get_qss_style
from ..._____assets import ASSETS
from qtmui.hooks import State
from ..utils.validate_params import _validate_param

class Rating(QFrame, PyWidgetBase):
    """
    A rating component, styled like Material-UI Rating.

    The `Rating` component allows users to select a rating value using icons. It integrates with the `qtmui`
    framework, retaining existing parameters, adding new parameters, and aligning with MUI Rating props.
    Inherits from native component props.

    Parameters
    ----------
    max : State or int, optional
        Maximum rating. Default is 5.
        Can be a `State` object for dynamic updates.
    value : State, float, or None, optional
        The rating value. Default is None.
        Can be a `State` object for dynamic updates.
    disabled : State or bool, optional
        If True, the component is disabled. Default is False.
        Can be a `State` object for dynamic updates.
    icon : State, QWidget, Callable, or None, optional
        The icon to display for selected ratings. Default is None.
        Can be a `State` object for dynamic updates or a Callable returning a QWidget.
    color : State or str, optional
        The color of the component. Default is None.
        Can be a `State` object for dynamic updates.
    name : State or str, optional
        The name attribute of the radio input elements. Default is None.
        Can be a `State` object for dynamic updates.
    readOnly : State or bool, optional
        If True, removes all hover effects and pointer events. Default is False.
        Can be a `State` object for dynamic updates.
    precision : State or float, optional
        The minimum increment value change allowed. Default is 1.
        Can be a `State` object for dynamic updates.
    onChange : State or Callable, optional
        Callback fired when the value changes. Default is None.
        Can be a `State` object for dynamic updates.
        Signature: (event: Any, value: float | None) -> None
    onChangeActive : State or Callable, optional
        Callback fired when the hover state changes. Default is None.
        Can be a `State` object for dynamic updates.
        Signature: (event: Any, value: float) -> None
    size : State or str, optional
        The size of the component ('small', 'medium', 'large'). Default is 'medium'.
        Can be a `State` object for dynamic updates.
    sx : State, list, dict, Callable, str, or None, optional
        System prop for CSS overrides and additional styles. Default is None.
        Can be a `State` object for dynamic updates.
    classes : State or dict, optional
        Override or extend the styles applied to the component. Default is None.
        Can be a `State` object for dynamic updates.
    component : State or type, optional
        The component used for the root node. Default is None (uses QFrame).
        Can be a `State` object for dynamic updates.
    defaultValue : State, float, or None, optional
        The default value when not controlled. Default is None.
        Can be a `State` object for dynamic updates.
    emptyIcon : State, QWidget, Callable, or None, optional
        The icon to display when empty. Default is None.
        Can be a `State` object for dynamic updates or a Callable returning a QWidget.
    emptyLabelText : State or str, optional
        The label read when the rating input is empty. Default is None.
        Can be a `State` object for dynamic updates.
    getLabelText : State or Callable, optional
        Function to provide a user-friendly name for the current value. Default is None.
        Can be a `State` object for dynamic updates.
        Signature: (value: float) -> str
    highlightSelectedOnly : State or bool, optional
        If True, only the selected icon is highlighted. Default is False.
        Can be a `State` object for dynamic updates.
    IconContainerComponent : State or type, optional
        The component containing the icon (deprecated, use slotProps.icon). Default is None.
        Can be a `State` object for dynamic updates.
    slotProps : State or dict, optional
        Props for slots ({decimal, icon, label, root}). Default is {}.
        Can be a `State` object for dynamic updates.
    slots : State or dict, optional
        Components for slots ({decimal, icon, label, root}). Default is {}.
        Can be a `State` object for dynamic updates.
    **kwargs
        Additional keyword arguments passed to the parent component,
        supporting props of the native component.

    Signals
    -------
    changed : Signal
        Emitted when the rating value changes.
    activeChanged : Signal
        Emitted when the hover state changes.

    Notes
    -----
    - Existing parameters are retained; new parameters added to align with MUI.
    - Props of the native component are supported via `**kwargs`.
    - MUI classes applied: `MuiRating-root`, `Mui-disabled`.
    - Integrates with `PyToolButton` and `PyIconify` for icon rendering.

    Demos:
    - Rating: https://qtmui.com/material-ui/qtmui-rating/

    API Reference:
    - Rating API: https://qtmui.com/material-ui/api/rating/
    """

    changed = Signal(object, float)
    activeChanged = Signal(object, float)

    VALID_SIZES = ['small', 'medium', 'large']

    def __init__(
        self,
        max: Union[State, int] = 5,
        value: Optional[Union[State, float]] = None,
        disabled: Union[State, bool] = False,
        icon: Optional[Union[State, QWidget, Callable]] = None,
        color: Optional[Union[State, str]] = None,
        name: Optional[Union[State, str]] = None,
        readOnly: Union[State, bool] = False,
        precision: Union[State, float] = 1,
        onChange: Optional[Union[State, Callable]] = None,
        onChangeActive: Optional[Union[State, Callable]] = None,
        size: Union[State, str] = 'medium',
        sx: Optional[Union[State, List, Dict, Callable, str]] = None,
        classes: Optional[Union[State, Dict]] = None,
        component: Optional[Union[State, type]] = None,
        defaultValue: Optional[Union[State, float]] = None,
        emptyIcon: Optional[Union[State, QWidget, Callable]] = None,
        emptyLabelText: Optional[Union[State, str]] = None,
        getLabelText: Optional[Union[State, Callable]] = None,
        highlightSelectedOnly: Union[State, bool] = False,
        IconContainerComponent: Optional[Union[State, type]] = None,
        slotProps: Union[State, Dict] = {},
        slots: Union[State, Dict] = {},
        *args,
        **kwargs
    ):
        root_component = component if component else QFrame
        super().__init__(*args, **kwargs, __class__=root_component)
        self.setObjectName(f"Rating-{str(uuid.uuid4())}")
        PyWidgetBase._setUpUi(self)

        self.theme = useTheme()
        self._widget_references = []
        self._currentValue = None

        # Set properties with validation
        self._set_max(max)
        self._set_value(value)
        self._set_disabled(disabled)
        self._set_icon(icon)
        self._set_color(color)
        self._set_name(name)
        self._set_readOnly(readOnly)
        self._set_precision(precision)
        self._set_onChange(onChange)
        self._set_onChangeActive(onChangeActive)
        self._set_size(size)
        self._set_sx(sx)
        self._set_classes(classes)
        self._set_component(component)
        self._set_defaultValue(defaultValue)
        self._set_emptyIcon(emptyIcon)
        self._set_emptyLabelText(emptyLabelText)
        self._set_getLabelText(getLabelText)
        self._set_highlightSelectedOnly(highlightSelectedOnly)
        self._set_IconContainerComponent(IconContainerComponent)
        self._set_slotProps(slotProps)
        self._set_slots(slots)

        self._init_ui()
        self._set_stylesheet()

        self.useEffect(
            self._set_stylesheet,
            [theme.state]
        )
        self.destroyed.connect(self._on_destroyed)
        self._connect_signals()

    # Setter and Getter methods
    @_validate_param(file_path="qtmui.material.rating", param_name="max", supported_signatures=Union[State, int], validator=lambda x: x > 0)
    def _set_max(self, value):
        """Assign value to max."""
        self._max = value

    def _get_max(self):
        """Get the max value."""
        return self._max.value if isinstance(self._max, State) else self._max

    @_validate_param(file_path="qtmui.material.rating", param_name="value", supported_signatures=Union[State, float, type(None)])
    def _set_value(self, value):
        """Assign value to value."""
        self._value = value
        self._currentValue = value.value if isinstance(value, State) else value

    def _get_value(self):
        """Get the value value."""
        return self._value.value if isinstance(self._value, State) else self._value

    @_validate_param(file_path="qtmui.material.rating", param_name="disabled", supported_signatures=Union[State, bool])
    def _set_disabled(self, value):
        """Assign value to disabled."""
        self._disabled = value
        self.setEnabled(not (value.value if isinstance(value, State) else value))

    def _get_disabled(self):
        """Get the disabled value."""
        return self._disabled.value if isinstance(self._disabled, State) else self._disabled

    @_validate_param(file_path="qtmui.material.rating", param_name="icon", supported_signatures=Union[State, QWidget, Callable, type(None)])
    def _set_icon(self, value):
        """Assign value to icon."""
        self._icon = value

    def _get_icon(self):
        """Get the icon value."""
        icon = self._icon
        if isinstance(icon, State):
            icon = icon.value
        if callable(icon):
            icon = icon()
        return icon if isinstance(icon, QWidget) else None

    @_validate_param(file_path="qtmui.material.rating", param_name="color", supported_signatures=Union[State, str, type(None)])
    def _set_color(self, value):
        """Assign value to color."""
        self._color = value

    def _get_color(self):
        """Get the color value."""
        return self._color.value if isinstance(self._color, State) else self._color

    @_validate_param(file_path="qtmui.material.rating", param_name="name", supported_signatures=Union[State, str, type(None)])
    def _set_name(self, value):
        """Assign value to name."""
        self._name = value

    def _get_name(self):
        """Get the name value."""
        return self._name.value if isinstance(self._name, State) else self._name

    @_validate_param(file_path="qtmui.material.rating", param_name="readOnly", supported_signatures=Union[State, bool])
    def _set_readOnly(self, value):
        """Assign value to readOnly."""
        self._readOnly = value

    def _get_readOnly(self):
        """Get the readOnly value."""
        return self._readOnly.value if isinstance(self._readOnly, State) else self._readOnly

    @_validate_param(file_path="qtmui.material.rating", param_name="precision", supported_signatures=Union[State, float], validator=lambda x: x > 0)
    def _set_precision(self, value):
        """Assign value to precision."""
        self._precision = value

    def _get_precision(self):
        """Get the precision value."""
        return self._precision.value if isinstance(self._precision, State) else self._precision

    @_validate_param(file_path="qtmui.material.rating", param_name="onChange", supported_signatures=Union[State, Callable, type(None)])
    def _set_onChange(self, value):
        """Assign value to onChange."""
        self._onChange = value

    def _get_onChange(self):
        """Get the onChange value."""
        return self._onChange.value if isinstance(self._onChange, State) else self._onChange

    @_validate_param(file_path="qtmui.material.rating", param_name="onChangeActive", supported_signatures=Union[State, Callable, type(None)])
    def _set_onChangeActive(self, value):
        """Assign value to onChangeActive."""
        self._onChangeActive = value

    def _get_onChangeActive(self):
        """Get the onChangeActive value."""
        return self._onChangeActive.value if isinstance(self._onChangeActive, State) else self._onChangeActive

    @_validate_param(file_path="qtmui.material.rating", param_name="size", supported_signatures=Union[State, str], valid_values=VALID_SIZES)
    def _set_size(self, value):
        """Assign value to size."""
        self._size = value

    def _get_size(self):
        """Get the size value."""
        return self._size.value if isinstance(self._size, State) else self._size

    @_validate_param(file_path="qtmui.material.rating", param_name="sx", supported_signatures=Union[State, List, Dict, Callable, str, type(None)])
    def _set_sx(self, value):
        """Assign value to sx."""
        self._sx = value

    def _get_sx(self):
        """Get the sx value."""
        return self._sx.value if isinstance(self._sx, State) else self._sx

    @_validate_param(file_path="qtmui.material.rating", param_name="classes", supported_signatures=Union[State, Dict, type(None)])
    def _set_classes(self, value):
        """Assign value to classes."""
        self._classes = value

    def _get_classes(self):
        """Get the classes value."""
        return self._classes.value if isinstance(self._classes, State) else self._classes

    @_validate_param(file_path="qtmui.material.rating", param_name="component", supported_signatures=Union[State, type, type(None)])
    def _set_component(self, value):
        """Assign value to component."""
        self._component = value

    def _get_component(self):
        """Get the component value."""
        return self._component.value if isinstance(self._component, State) else self._component

    @_validate_param(file_path="qtmui.material.rating", param_name="defaultValue", supported_signatures=Union[State, float, type(None)])
    def _set_defaultValue(self, value):
        """Assign value to defaultValue."""
        self._defaultValue = value

    def _get_defaultValue(self):
        """Get the defaultValue value."""
        return self._defaultValue.value if isinstance(self._defaultValue, State) else self._defaultValue

    @_validate_param(file_path="qtmui.material.rating", param_name="emptyIcon", supported_signatures=Union[State, QWidget, Callable, type(None)])
    def _set_emptyIcon(self, value):
        """Assign value to emptyIcon."""
        self._emptyIcon = value

    def _get_emptyIcon(self):
        """Get the emptyIcon value."""
        icon = self._emptyIcon
        if isinstance(icon, State):
            icon = icon.value
        if callable(icon):
            icon = icon()
        return icon if isinstance(icon, QWidget) else None

    @_validate_param(file_path="qtmui.material.rating", param_name="emptyLabelText", supported_signatures=Union[State, str, type(None)])
    def _set_emptyLabelText(self, value):
        """Assign value to emptyLabelText."""
        self._emptyLabelText = value

    def _get_emptyLabelText(self):
        """Get the emptyLabelText value."""
        return self._emptyLabelText.value if isinstance(self._emptyLabelText, State) else self._emptyLabelText

    @_validate_param(file_path="qtmui.material.rating", param_name="getLabelText", supported_signatures=Union[State, Callable, type(None)])
    def _set_getLabelText(self, value):
        """Assign value to getLabelText."""
        self._getLabelText = value

    def _get_getLabelText(self):
        """Get the getLabelText value."""
        return self._getLabelText.value if isinstance(self._getLabelText, State) else self._getLabelText

    @_validate_param(file_path="qtmui.material.rating", param_name="highlightSelectedOnly", supported_signatures=Union[State, bool])
    def _set_highlightSelectedOnly(self, value):
        """Assign value to highlightSelectedOnly."""
        self._highlightSelectedOnly = value

    def _get_highlightSelectedOnly(self):
        """Get the highlightSelectedOnly value."""
        return self._highlightSelectedOnly.value if isinstance(self._highlightSelectedOnly, State) else self._highlightSelectedOnly

    @_validate_param(file_path="qtmui.material.rating", param_name="IconContainerComponent", supported_signatures=Union[State, type, type(None)])
    def _set_IconContainerComponent(self, value):
        """Assign value to IconContainerComponent."""
        self._IconContainerComponent = value

    def _get_IconContainerComponent(self):
        """Get the IconContainerComponent value."""
        return self._IconContainerComponent.value if isinstance(self._IconContainerComponent, State) else self._IconContainerComponent

    @_validate_param(file_path="qtmui.material.rating", param_name="slotProps", supported_signatures=Union[State, Dict])
    def _set_slotProps(self, value):
        """Assign value to slotProps."""
        self._slotProps = value

    def _get_slotProps(self):
        """Get the slotProps value."""
        return self._slotProps.value if isinstance(self._slotProps, State) else self._slotProps

    @_validate_param(file_path="qtmui.material.rating", param_name="slots", supported_signatures=Union[State, Dict])
    def _set_slots(self, value):
        """Assign value to slots."""
        self._slots = value

    def _get_slots(self):
        """Get the slots value."""
        return self._slots.value if isinstance(self._slots, State) else self._slots

    def _init_ui(self):
        """Initialize the UI based on props."""
        self.theme = useTheme()
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)

        # Clear previous widgets
        self._widget_references.clear()
        while self.layout().count():
            item = self.layout().takeAt(0)
            if item.widget():
                item.widget().setParent(None)

        # Determine initial value
        initial_value = self._get_value() if self._get_value() is not None else self._get_defaultValue()
        self._currentValue = initial_value

        # Get icon sizes from theme
        icon_styles = self.theme.components.get("Rating", {}).get("styles", {}).get("root", {}).get("props", {})
        size_key = f"size{self._get_size().capitalize()}"
        icon_width = icon_styles.get(size_key, {}).get("svgIcon", {}).get("width", 24)
        icon_height = icon_styles.get(size_key, {}).get("svgIcon", {}).get("height", 24)
        self._iconSize = QSize(icon_width, icon_height)

        # Colors
        self._selected_color = self.theme.palette.primary.main
        self._unselected_color = self.theme.palette.text.secondary

        # Create rating buttons
        self._icons = []
        max_rating = self._get_max()
        for i in range(max_rating):
            icon_widget = self._get_icon() or PyIconify(key=ASSETS.ICONS.STAR, color=self._unselected_color, size=self._iconSize)
            empty_icon_widget = self._get_emptyIcon() or PyIconify(key=ASSETS.ICONS.STAR_OUTLINE, color=self._unselected_color, size=self._iconSize)
            container = self._get_IconContainerComponent() or PyToolButton
            rating_button = container(
                icon=icon_widget if i < (self._currentValue or 0) else empty_icon_widget,
                size=self._iconSize,
                tooltip=self._get_label_text(i + 1)
            )
            rating_button.setCursor(QCursor(Qt.PointingHandCursor) if not (self._get_disabled() or self._get_readOnly()) else QCursor(Qt.ArrowCursor))
            rating_button.setIconSize(self._iconSize)

            # Set initial color
            if i < (self._currentValue or 0):
                rating_button._set_text_color(self._selected_color)
            else:
                rating_button._set_text_color(self._unselected_color)

            # Connect events
            rating_button.enterEvent = lambda event, idx=i: self._on_hover_enter(idx)
            rating_button.leaveEvent = lambda event, idx=i: self._on_hover_leave(idx)
            rating_button.clicked.connect(lambda checked, idx=i: self._on_icon_clicked(idx))

            self._layout.addWidget(rating_button)
            self._icons.append(rating_button)
            self._widget_references.append(rating_button)

        self.setFixedWidth(self.sizeHint().width() + 10)
        self.setMinimumHeight(self.sizeHint().height() + 5)

    def _get_label_text(self, value):
        """Get the label text for a given value."""
        get_label_text = self._get_getLabelText()
        if get_label_text:
            return get_label_text(value)
        return f"{value or '0'} Star{'s' if value != 1 else ''}" if value else self._get_emptyLabelText() or 'Empty'

    def _on_hover_enter(self, index):
        """Handle hover enter event."""
        if self._get_disabled() or self._get_readOnly():
            return
        highlight_only = self._get_highlightSelectedOnly()
        for i in range(self._get_max()):
            if highlight_only and i != index:
                self._icons[i]._set_text_color(self._unselected_color)
                self._icons[i].setIconSize(self._iconSize)
                self._icons[i].setFixedSize(self._iconSize.width(), self._iconSize.height())
            elif i <= index:
                self._icons[i]._set_text_color(self._selected_color)
                self._icons[i].setIconSize(QSize(self._iconSize.width() + 5, self._iconSize.height() + 5))
                self._icons[i].setFixedSize(self._iconSize.width() + 5, self._iconSize.height() + 5)
            else:
                self._icons[i]._set_text_color(self._unselected_color)
                self._icons[i].setIconSize(self._iconSize)
                self._icons[i].setFixedSize(self._iconSize.width(), self._iconSize.height())
        if self._get_onChangeActive():
            self._get_onChangeActive()(None, index + 1)
            self.activeChanged.emit(None, index + 1)

    def _on_hover_leave(self, index):
        """Handle hover leave event."""
        if self._get_disabled() or self._get_readOnly():
            return
        for i in range(self._get_max()):
            if i < (self._currentValue or 0):
                self._icons[i]._set_text_color(self._selected_color)
            else:
                self._icons[i]._set_text_color(self._unselected_color)
            self._icons[i].setIconSize(self._iconSize)
            self._icons[i].setFixedSize(self._iconSize.width(), self._iconSize.height())

    def _on_icon_clicked(self, index):
        """Handle icon click event."""
        if self._get_disabled() or self._get_readOnly():
            return
        new_value = index + 1
        precision = self._get_precision()
        new_value = round(new_value / precision) * precision
        self._currentValue = new_value
        self._set_value(new_value)
        self._on_hover_enter(index)
        self.changed.emit(None, new_value)
        if self._get_onChange():
            self._get_onChange()(None, new_value)

    def _set_stylesheet(self, component_styled=None):
        """Set the stylesheet for the Rating."""
        self.theme = useTheme()
        component_styled = component_styled or self.theme.components
        rating_styles = component_styled.get("Rating", {}).get("styles", {})
        root_styles = rating_styles.get("root", {})
        root_qss = get_qss_style(root_styles)

        # Handle sx
        sx = self._get_sx()
        sx_qss = ""
        if sx:
            if isinstance(sx, (list, dict)):
                sx_qss = get_qss_style(sx, class_name=f"#{self.objectName()}")
            elif isinstance(sx, Callable):
                sx_result = sx()
                if isinstance(sx_result, (list, dict)):
                    sx_qss = get_qss_style(sx_result, class_name=f"#{self.objectName()}")
                elif isinstance(sx_result, str):
                    sx_qss = sx_result
            elif isinstance(sx, str) and sx != "":
                sx_qss = sx

        # Handle classes
        classes = self._get_classes()
        classes_qss = get_qss_style(classes, class_name=f"#{self.objectName()}") if classes else ""

        # Handle slotProps.root
        root_props = self._get_slotProps().get('root', {})
        root_props_qss = get_qss_style(root_props.get('sx', {}), class_name=f"#{self.objectName()}")

        # Apply MUI classes
        mui_classes = ["MuiRating-root"]
        if self._get_disabled():
            mui_classes.append("Mui-disabled")

        stylesheet = f"""
            #{self.objectName()} {{
                {root_qss}
                {classes_qss}
                {root_props_qss}
                background: transparent;
            }}
            {sx_qss}
        """
        self.setStyleSheet(stylesheet)

    def _connect_signals(self):
        """Connect valueChanged signals of State parameters to their slots."""
        if isinstance(self._max, State):
            self._max.valueChanged.connect(self._on_max_changed)
        if isinstance(self._value, State):
            self._value.valueChanged.connect(self._on_value_changed)
        if isinstance(self._disabled, State):
            self._disabled.valueChanged.connect(self._on_disabled_changed)
        if isinstance(self._icon, State):
            self._icon.valueChanged.connect(self._on_icon_changed)
        if isinstance(self._color, State):
            self._color.valueChanged.connect(self._on_color_changed)
        if isinstance(self._name, State):
            self._name.valueChanged.connect(self._on_name_changed)
        if isinstance(self._readOnly, State):
            self._readOnly.valueChanged.connect(self._on_readOnly_changed)
        if isinstance(self._precision, State):
            self._precision.valueChanged.connect(self._on_precision_changed)
        if isinstance(self._onChange, State):
            self._onChange.valueChanged.connect(self._on_onChange_changed)
        if isinstance(self._onChangeActive, State):
            self._onChangeActive.valueChanged.connect(self._on_onChangeActive_changed)
        if isinstance(self._size, State):
            self._size.valueChanged.connect(self._on_size_changed)
        if isinstance(self._sx, State):
            self._sx.valueChanged.connect(self._on_sx_changed)
        if isinstance(self._classes, State):
            self._classes.valueChanged.connect(self._on_classes_changed)
        if isinstance(self._component, State):
            self._component.valueChanged.connect(self._on_component_changed)
        if isinstance(self._defaultValue, State):
            self._defaultValue.valueChanged.connect(self._on_defaultValue_changed)
        if isinstance(self._emptyIcon, State):
            self._emptyIcon.valueChanged.connect(self._on_emptyIcon_changed)
        if isinstance(self._emptyLabelText, State):
            self._emptyLabelText.valueChanged.connect(self._on_emptyLabelText_changed)
        if isinstance(self._getLabelText, State):
            self._getLabelText.valueChanged.connect(self._on_getLabelText_changed)
        if isinstance(self._highlightSelectedOnly, State):
            self._highlightSelectedOnly.valueChanged.connect(self._on_highlightSelectedOnly_changed)
        if isinstance(self._IconContainerComponent, State):
            self._IconContainerComponent.valueChanged.connect(self._on_IconContainerComponent_changed)
        if isinstance(self._slotProps, State):
            self._slotProps.valueChanged.connect(self._on_slotProps_changed)
        if isinstance(self._slots, State):
            self._slots.valueChanged.connect(self._on_slots_changed)

    def _on_max_changed(self):
        """Handle changes to max."""
        self._set_max(self._max)
        self._init_ui()

    def _on_value_changed(self):
        """Handle changes to value."""
        self._set_value(self._value)
        self._init_ui()

    def _on_disabled_changed(self):
        """Handle changes to disabled."""
        self._set_disabled(self._disabled)
        self._set_stylesheet()

    def _on_icon_changed(self):
        """Handle changes to icon."""
        self._set_icon(self._icon)
        self._init_ui()

    def _on_color_changed(self):
        """Handle changes to color."""
        self._set_color(self._color)
        self._set_stylesheet()

    def _on_name_changed(self):
        """Handle changes to name."""
        self._set_name(self._name)

    def _on_readOnly_changed(self):
        """Handle changes to readOnly."""
        self._set_readOnly(self._readOnly)
        self._init_ui()

    def _on_precision_changed(self):
        """Handle changes to precision."""
        self._set_precision(self._precision)
        self._init_ui()

    def _on_onChange_changed(self):
        """Handle changes to onChange."""
        self._set_onChange(self._onChange)

    def _on_onChangeActive_changed(self):
        """Handle changes to onChangeActive."""
        self._set_onChangeActive(self._onChangeActive)

    def _on_size_changed(self):
        """Handle changes to size."""
        self._set_size(self._size)
        self._init_ui()

    def _on_sx_changed(self):
        """Handle changes to sx."""
        self._set_sx(self._sx)
        self._set_stylesheet()

    def _on_classes_changed(self):
        """Handle changes to classes."""
        self._set_classes(self._classes)
        self._set_stylesheet()

    def _on_component_changed(self):
        """Handle changes to component."""
        self._set_component(self._component)
        self._init_ui()

    def _on_defaultValue_changed(self):
        """Handle changes to defaultValue."""
        self._set_defaultValue(self._defaultValue)
        self._init_ui()

    def _on_emptyIcon_changed(self):
        """Handle changes to emptyIcon."""
        self._set_emptyIcon(self._emptyIcon)
        self._init_ui()

    def _on_emptyLabelText_changed(self):
        """Handle changes to emptyLabelText."""
        self._set_emptyLabelText(self._emptyLabelText)
        self._init_ui()

    def _on_getLabelText_changed(self):
        """Handle changes to getLabelText."""
        self._set_getLabelText(self._getLabelText)
        self._init_ui()

    def _on_highlightSelectedOnly_changed(self):
        """Handle changes to highlightSelectedOnly."""
        self._set_highlightSelectedOnly(self._highlightSelectedOnly)
        self._init_ui()

    def _on_IconContainerComponent_changed(self):
        """Handle changes to IconContainerComponent."""
        self._set_IconContainerComponent(self._IconContainerComponent)
        self._init_ui()

    def _on_slotProps_changed(self):
        """Handle changes to slotProps."""
        self._set_slotProps(self._slotProps)
        self._set_stylesheet()

    def _on_slots_changed(self):
        """Handle changes to slots."""
        self._set_slots(self._slots)
        self._init_ui()

    def _on_destroyed(self):
        """Clean up connections when the widget is destroyed."""
        if hasattr(self, "theme"):
            self.theme.state.valueChanged.disconnect(self._set_stylesheet)
        if isinstance(self._max, State):
            self._max.valueChanged.disconnect(self._on_max_changed)
        if isinstance(self._value, State):
            self._value.valueChanged.disconnect(self._on_value_changed)
        if isinstance(self._disabled, State):
            self._disabled.valueChanged.disconnect(self._on_disabled_changed)
        if isinstance(self._icon, State):
            self._icon.valueChanged.disconnect(self._on_icon_changed)
        if isinstance(self._color, State):
            self._color.valueChanged.disconnect(self._on_color_changed)
        if isinstance(self._name, State):
            self._name.valueChanged.disconnect(self._on_name_changed)
        if isinstance(self._readOnly, State):
            self._readOnly.valueChanged.disconnect(self._on_readOnly_changed)
        if isinstance(self._precision, State):
            self._precision.valueChanged.disconnect(self._on_precision_changed)
        if isinstance(self._onChange, State):
            self._onChange.valueChanged.disconnect(self._on_onChange_changed)
        if isinstance(self._onChangeActive, State):
            self._onChangeActive.valueChanged.disconnect(self._on_onChangeActive_changed)
        if isinstance(self._size, State):
            self._size.valueChanged.disconnect(self._on_size_changed)
        if isinstance(self._sx, State):
            self._sx.valueChanged.disconnect(self._on_sx_changed)
        if isinstance(self._classes, State):
            self._classes.valueChanged.disconnect(self._on_classes_changed)
        if isinstance(self._component, State):
            self._component.valueChanged.disconnect(self._on_component_changed)
        if isinstance(self._defaultValue, State):
            self._defaultValue.valueChanged.disconnect(self._on_defaultValue_changed)
        if isinstance(self._emptyIcon, State):
            self._emptyIcon.valueChanged.disconnect(self._on_emptyIcon_changed)
        if isinstance(self._emptyLabelText, State):
            self._emptyLabelText.valueChanged.disconnect(self._on_emptyLabelText_changed)
        if isinstance(self._getLabelText, State):
            self._getLabelText.valueChanged.disconnect(self._on_getLabelText_changed)
        if isinstance(self._highlightSelectedOnly, State):
            self._highlightSelectedOnly.valueChanged.disconnect(self._on_highlightSelectedOnly_changed)
        if isinstance(self._IconContainerComponent, State):
            self._IconContainerComponent.valueChanged.disconnect(self._on_IconContainerComponent_changed)
        if isinstance(self._slotProps, State):
            self._slotProps.valueChanged.disconnect(self._on_slotProps_changed)
        if isinstance(self._slots, State):
            self._slots.valueChanged.disconnect(self._on_slots_changed)


from typing import Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from qtmui.material.styles.create_theme.theme_reducer import ThemeState

from .properties_name import *
from ....system.color_manipulator import alpha


def py_tool_button(_theme) -> Dict:
    theme: ThemeState = _theme

    COLORS = ['primary', 'secondary', 'info', 'success', 'warning', 'error']
    lightMode = theme.palette.mode == 'light'

    return {
        'PyToolButton': {
            'styles': {
                'root':  {
                    **{
                        "default": {
                            backgroundColor: "transparent",
                            color: theme.palette.common.white if lightMode else theme.palette.grey._800,
                            "slots": {
                                "hover": {
                                    color: theme.palette.action.hover,
                                }
                            },
                        },
                    },
                    **{
                        f"{_color}": {
                            backgroundColor: "transparent",
                            color: getattr(theme.palette, _color).main,
                            "slots": {
                                "hover": {
                                    color: alpha(getattr(theme.palette, _color).main, 0.08),
                                },
                            },
                        }
                        for _color in COLORS
                    }
                }
            }
        },
        'PyToolButtonSize': {
            'styles': {
                'small': {
                    height: '20px',
                    borderRadius: "10px"
                },
                'medium': {
                    height: '24px',
                    borderRadius: "12px"
                },
            }
        }
    }



from typing import Optional

class NavConfigProps:
    def __init__(self,
                 hiddenLabel: Optional[bool] = False,
                 itemGap: Optional[int] = None,
                 iconSize: Optional[int] = None,
                 itemRadius: Optional[int] = None,
                 itemPadding: Optional[str] = None,
                 currentRole: Optional[str] = None,
                 itemSubHeight: Optional[int] = None,
                 itemRootHeight: Optional[int] = None):
        self.hiddenLabel = hiddenLabel
        self.itemGap = itemGap
        self.iconSize = iconSize
        self.itemRadius = itemRadius
        self.itemPadding = itemPadding
        self.currentRole = currentRole
        self.itemSubHeight = itemSubHeight
        self.itemRootHeight = itemRootHeight

def navVerticalConfig(config: Optional[NavConfigProps] = None) -> dict:
    config = config or NavConfigProps()
    return {
        'itemGap': config.itemGap if config.itemGap is not None else 4,
        'iconSize': config.iconSize if config.iconSize is not None else 24,
        'currentRole': config.currentRole,
        'itemRootHeight': config.itemRootHeight if config.itemRootHeight is not None else 44,
        'itemSubHeight': config.itemSubHeight if config.itemSubHeight is not None else 36,
        'itemPadding': config.itemPadding if config.itemPadding is not None else '4px 8px 4px 12px',
        'itemRadius': config.itemRadius if config.itemRadius is not None else 8,
        'hiddenLabel': config.hiddenLabel if config.hiddenLabel is not None else False,
    }

def navMiniConfig(config: Optional[NavConfigProps] = None) -> dict:
    config = config or NavConfigProps()
    return {
        'itemGap': config.itemGap if config.itemGap is not None else 4,
        'iconSize': config.iconSize if config.iconSize is not None else 22,
        'currentRole': config.currentRole,
        'itemRootHeight': config.itemRootHeight if config.itemRootHeight is not None else 56,
        'itemSubHeight': config.itemSubHeight if config.itemSubHeight is not None else 34,
        'itemPadding': config.itemPadding if config.itemPadding is not None else '6px 0 0 0',
        'itemRadius': config.itemRadius if config.itemRadius is not None else 6,
        'hiddenLabel': config.hiddenLabel if config.hiddenLabel is not None else False,
    }

def navHorizontalConfig(config: Optional[NavConfigProps] = None) -> dict:
    config = config or NavConfigProps()
    return {
        'itemGap': config.itemGap if config.itemGap is not None else 6,
        'iconSize': config.iconSize if config.iconSize is not None else 22,
        'currentRole': config.currentRole,
        'itemRootHeight': config.itemRootHeight if config.itemRootHeight is not None else 32,
        'itemSubHeight': config.itemSubHeight if config.itemSubHeight is not None else 34,
        'itemPadding': config.itemPadding if config.itemPadding is not None else '0 6px 0 6px',
        'itemRadius': config.itemRadius if config.itemRadius is not None else 6,
        'hiddenLabel': config.hiddenLabel if config.hiddenLabel is not None else False,
    }

# # Ví dụ sử dụng các hàm
# nav_config = NavConfigProps(itemGap=10, iconSize=30)
# print(navVerticalConfig(nav_config))
# print(navMiniConfig(nav_config))
# print(navHorizontalConfig(nav_config))



import sys

# ============================================================
# Qt imports
# ============================================================
from PySide6.QtWidgets import (
    QApplication,
    QWidget,
    QPushButton,
    QVBoxLayout,
    QGraphicsEffect,
)
from PySide6.QtCore import (
    Qt,
    QRectF,
    Property,
    QPropertyAnimation,
)
from PySide6.QtGui import (
    QPainter,
    QColor,
    QPainterPath,
    QConicalGradient,
)

# ============================================================
# AnimEffect
# ------------------------------------------------------------
# Đây là unified effect cho qtmui:
# - KHÔNG phụ thuộc widget cụ thể
# - Vẽ gradient border bằng kỹ thuật giống CSS ::before
# - Không làm hỏng layout, không clip sai
#
# Triết lý:
#   CSS:
#     element
#       ├── ::before (gradient border)
#       └── content
#
#   Qt:
#     QGraphicsEffect.draw()
#       ├── draw gradient layer (ngoài)
#       ├── clip inner rounded rect
#       └── drawSource() (widget gốc)
# ============================================================
class AnimEffect(QGraphicsEffect):
    def __init__(self, parent=None):
        super().__init__(parent)

        # =====================================================
        # Animation state
        # -----------------------------------------------------
        # angle: dùng cho QConicalGradient
        # Qt sẽ gọi setAngle() liên tục khi animation chạy
        # =====================================================
        self._angle = 0.0

        # =====================================================
        # Border config
        # -----------------------------------------------------
        # border : độ dày viền gradient
        # radius : bo góc content (PHẢI khớp với QPushButton)
        # =====================================================
        self.border = 2
        self.radius = 28

        # =====================================================
        # Gradient colors (cyclic)
        # -----------------------------------------------------
        # Màu đầu = màu cuối để loop liền mạch
        # =====================================================
        self.colors = [
            QColor("#ff4545"),
            QColor("#00ff99"),
            QColor("#006aff"),
            QColor("#ff0095"),
            QColor("#ff4545"),
        ]

    # =====================================================
    # boundingRectFor
    # -----------------------------------------------------
    # BẮT BUỘC override
    #
    # Lý do:
    # - QGraphicsEffect mặc định clip theo rect widget
    # - Border nằm NGOÀI rect → sẽ bị cắt
    #
    # Giải pháp:
    # - Mở rộng rect thêm border px mỗi phía
    # =====================================================
    def boundingRectFor(self, rect: QRectF) -> QRectF:
        b = self.border
        return rect.adjusted(-b, -b, b, b)

    # =====================================================
    # angle property
    # -----------------------------------------------------
    # Dùng cho QPropertyAnimation
    # Mỗi lần set → update() → draw()
    # =====================================================
    def getAngle(self):
        return self._angle

    def setAngle(self, v):
        self._angle = v
        self.update()

    angle = Property(float, getAngle, setAngle)

    # =====================================================
    # draw()
    # -----------------------------------------------------
    # THỨ TỰ VẼ RẤT QUAN TRỌNG:
    #
    # 1. Vẽ gradient border (outer)
    # 2. Clip inner rounded rect
    # 3. drawSource() → widget gốc
    #
    # TUYỆT ĐỐI:
    # - Không dùng drawPixmap cho content
    # - drawSource() để tránh clip bug Qt
    # =====================================================
    def draw(self, painter: QPainter):
        b = self.border

        # rect của widget gốc (chưa expand)
        src_rect = self.sourceBoundingRect()
        w, h = src_rect.width(), src_rect.height()

        # tâm để xoay gradient
        cx, cy = w / 2, h / 2

        painter.save()
        painter.setRenderHint(QPainter.Antialiasing)

        # =================================================
        # 1. GRADIENT BORDER (CSS ::before equivalent)
        # -------------------------------------------------
        # Dùng QConicalGradient:
        # - center = tâm widget
        # - angle = self._angle (animate)
        #
        # Giống:
        # conic-gradient(from var(--angle), ...)
        # =================================================
        grad = QConicalGradient(cx, cy, self._angle)

        step = 1 / (len(self.colors) - 1)
        for i, c in enumerate(self.colors):
            grad.setColorAt(i * step, c)

        # Rect ngoài = widget + border
        outer_rect = QRectF(
            src_rect.left() - b,
            src_rect.top() - b,
            w + b * 2,
            h + b * 2,
        )

        # Rounded rect ngoài
        outer_path = QPainterPath()
        outer_path.addRoundedRect(
            outer_rect,
            self.radius + b,   # bo góc đồng tâm
            self.radius + b,
        )

        painter.fillPath(outer_path, grad)

        # =================================================
        # 2. CLIP INNER (mask) không có không sao
        # -------------------------------------------------
        # Cắt phần content để:
        # - Border KHÔNG đè lên content
        # - Bo góc đồng tâm hoàn hảo
        #
        # Đây chính là "dao cắt" giống CSS padding + mask
        # =================================================
        # inner_path = QPainterPath()
        # inner_path.addRoundedRect(
        #     src_rect,
        #     self.radius,
        #     self.radius,
        # )

        # painter.setClipPath(inner_path)

        # =================================================
        # 3. DRAW CONTENT
        # -------------------------------------------------
        # drawSource():
        # - Qt tự render widget gốc
        # - Không bị clip lệch phải / dưới
        # - Đúng tuyệt đối layout
        # =================================================
        self.drawSource(painter)

        painter.restore()


# ============================================================
# GradientBorderButton
# ------------------------------------------------------------
# Button bình thường + gắn AnimEffect
# Không custom paint button → giữ native behavior
# ============================================================
class GradientBorderButton(QPushButton):
    def __init__(self, text):
        super().__init__(text)

        # Kích thước demo
        self.setFixedSize(240, 56)

        # Style content
        self.setStyleSheet(
            """
            QPushButton {
                background-color: #121212;
                color: white;
                border: none;
                border-radius: 28px; /* PHẢI khớp AnimEffect.radius */
                font-size: 14px;
                font-weight: 600;
            }
            """
        )
        # self.setStyleSheet(
        #     """
        #     QPushButton {
        #         background-color: transparent;
        #         color: white;
        #         border: none;
        #         border-radius: 28px; /* PHẢI khớp AnimEffect.radius */
        #         font-size: 14px;
        #         font-weight: 600;
        #     }
        #     """
        # )

        # Gắn effect
        self.effect = AnimEffect(self)
        self.setGraphicsEffect(self.effect)

        # Animate gradient angle
        self.anim = QPropertyAnimation(self.effect, b"angle", self)
        self.anim.setStartValue(0)
        self.anim.setEndValue(360)
        self.anim.setDuration(3000)
        self.anim.setLoopCount(-1)
        self.anim.start()


# ============================================================
# Demo app
# ============================================================
class Demo(QWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("qtmui – Gradient Border Button (Documented)")
        self.resize(420, 300)

        layout = QVBoxLayout(self)
        layout.setAlignment(Qt.AlignCenter)
        layout.setSpacing(0)

        layout.addWidget(GradientBorderButton("Primary Action"))
        layout.addWidget(GradientBorderButton("Confirm"))


# ============================================================
# Entry
# ============================================================
if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = Demo()
    w.show()
    sys.exit(app.exec())


"""
tôi muốn xây dựng cơ chế after, before
""" 


import asyncio
from typing import Callable, Dict, List, Optional, Union

from qtmui.hooks import State
from PySide6.QtWidgets import QWidget
from PySide6.QtCore import QTimer

from qtmui.material.styles import useTheme

from .button import Button
from qtmui.material.styles.create_theme.components.get_qss_styles import get_qss_style

class Fab(Button):
    """
    A Floating Action Button (FAB) component based on Material-UI's Fab.

    The `Fab` component extends the `Button` component to provide a floating action button
    with circular or extended styling, supporting all props of the Material-UI `Fab` component,
    as well as additional custom props.

    Parameters
    ----------
    icon : State, str, or None, optional
        The icon to display (custom feature, not part of Material-UI). Maps to `startIcon`.
        Default is None.
        Can be a `State` object for dynamic updates.
    animate : State or bool, optional
        If True, applies a rotation animation (custom feature, not part of Material-UI).
        Default is False.
        Can be a `State` object for dynamic updates.
    children : State, QWidget, List[QWidget], or None, optional
        The content of the component, typically an icon or text. Default is None.
        Can be a `State` object for dynamic updates.
    classes : State or dict, optional
        Override or extend the styles applied to the component. Default is None.
        Can be a `State` object for dynamic updates.
    color : State or str, optional
        The color of the component ("default", "error", "info", "inherit", "primary",
        "secondary", "success", "warning", or custom string). Default is "default".
        Can be a `State` object for dynamic updates.
    component : State, str, or None, optional
        The component used for the root node (e.g., "QPushButton"). Default is None.
        Can be a `State` object for dynamic updates.
    disabled : State or bool, optional
        If True, the component is disabled. Default is False.
        Can be a `State` object for dynamic updates.
    disableFocusRipple : State or bool, optional
        If True, the keyboard focus ripple is disabled. Default is False.
        Can be a `State` object for dynamic updates.
    disableRipple : State or bool, optional
        If True, the ripple effect is disabled. Default is False.
        Can be a `State` object for dynamic updates.
    href : State, str, or None, optional
        The URL to link to when the button is clicked. Default is None.
        Can be a `State` object for dynamic updates.
    size : State or str, optional
        The size of the component ("small", "medium", "large", or custom string).
        Default is "large".
        Can be a `State` object for dynamic updates.
    sx : State, dict, Callable, str, or None, optional
        System prop for CSS overrides and additional styles. Default is None.
        Can be a `State` object for dynamic updates.
    variant : State or str, optional
        The variant to use ("circular", "extended", or custom string). Default is "circular".
        Can be a `State` object for dynamic updates.
    *args
        Additional positional arguments passed to the parent `Button` class.
    **kwargs
        Additional keyword arguments passed to the parent `Button` class, supporting
        props of the `ButtonBase` component (e.g., onClick, style).

    Attributes
    ----------
    VALID_COLORS : list[str]
        Valid values for `color`: ["default", "error", "info", "inherit", "primary",
        "secondary", "success", "warning"].
    VALID_SIZES : list[str]
        Valid values for `size`: ["small", "medium", "large"].
    VALID_VARIANTS : list[str]
        Valid values for `variant`: ["circular", "extended"].

    Notes
    -----
    - Props of the `ButtonBase` component are supported via `**kwargs` (e.g., `onClick`, `style`).
    - The `icon` and `animate` parameters are custom features, not part of Material-UI's `Fab`.
    - The `children` prop must be a `QWidget`, a list of `QWidget` instances, or a `State` object.

    Demos:
    - Fab: https://qtmui.com/material-ui/qtmui-fab/

    API Reference:
    - Fab API: https://qtmui.com/material-ui/api/fab/
    """

    def __init__(
                self,
                icon: Optional[Union[State, str]] = None,
                animate: Union[State, bool] = False,
                # children: Optional[Union[State, QWidget, List[QWidget]]] = None,
                # classes: Optional[Union[State, Dict]] = None,
                # color: Union[State, str] = "default",
                # component: Optional[Union[State, str]] = None,
                # disabled: Union[State, bool] = False,
                # disableFocusRipple: Union[State, bool] = False,
                # disableRipple: Union[State, bool] = False,
                # href: Optional[Union[State, str]] = None,
                # size: Union[State, str] = "large",
                sx: Optional[Union[State, Dict, Callable, str]] = None,
                variant: Union[State, str] = "contained",
                *args,
                **kwargs
                ):
        super().__init__(startIcon=icon, variant=variant, componentType="Fab", *args, **kwargs)

        self._sx = sx
        self._kwargs = kwargs
        self._animate = animate



# from typing import Optional, Type, TypeVar, Dict, Any

# class ListItemButtonBaseProps:
#     def __init__(
#         self,
#         alignItems: Optional[str] = 'center',
#         autoFocus: Optional[bool] = False,
#         children: Optional[Any] = None,
#         classes: Optional[Dict[str, Any]] = None,
#         dense: Optional[bool] = False,
#         disabled: Optional[bool] = False,
#         disableGutters: Optional[bool] = False,
#         divider: Optional[bool] = False,
#         selected: Optional[bool] = False,
#         sx: Optional[Any] = None
#     ):
#         self.alignItems = alignItems
#         self.autoFocus = autoFocus
#         self.children = children
#         self.classes = classes
#         self.dense = dense
#         self.disabled = disabled
#         self.disableGutters = disableGutters
#         self.divider = divider
#         self.selected = selected
#         self.sx = sx

#     def __repr__(self):
#         return (
#             f"ListItemButtonBaseProps(alignItems={self.alignItems}, autoFocus={self.autoFocus}, "
#             f"children={self.children}, classes={self.classes}, dense={self.dense}, disabled={self.disabled}, "
#             f"disableGutters={self.disableGutters}, divider={self.divider}, selected={self.selected}, sx={self.sx})"
#         )


# # Định nghĩa lớp ListItemButtonTypeMap kế thừa từ ExtendButtonBaseTypeMap
# class ListItemButtonTypeMap(ExtendButtonBaseTypeMap):
#     def __init__(
#         self,
#         props: Optional[Dict[str, Any]] = None,
#         defaultComponent: Optional[Any] = 'div'
#     ):
#         combined_props = {**props, **vars(ListItemButtonBaseProps())} if props else vars(ListItemButtonBaseProps())
#         super().__init__(props=combined_props, defaultComponent=defaultComponent)

#     def __repr__(self):
#         return f"ListItemButtonTypeMap(props={self.props}, defaultComponent={self.defaultComponent})"


# # Giả định `OverrideProps` là một lớp cơ sở với các thuộc tính cần thiết
# class OverrideProps:
#     def __init__(self, **kwargs):
#         for key, value in kwargs.items():
#             setattr(self, key, value)

# # Sử dụng generic type trong Python bằng cách sử dụng `TypeVar`
# RootComponent = TypeVar('RootComponent')

# # Định nghĩa lớp `ListItemButtonProps` kế thừa từ `ListItemButtonTypeMap` và `OverrideProps`
# class ListItemButtonProps(ListItemButtonTypeMap, OverrideProps):
#     def __init__(self,
#                  component: Optional[Type] = None,
#                  additional_props: Optional[Dict[str, Any]] = None,
#                  default_component: Optional[Any] = None,
#                  **kwargs):
#         ListItemButtonTypeMap.__init__(self, additional_props=additional_props, default_component=default_component)
#         OverrideProps.__init__(self, **kwargs)
#         self.component = component

#     def __repr__(self):
#         return (f"ListItemButtonProps(component={self.component}, additional_props={self.additional_props}, "
#                 f"default_component={self.default_component}, other_props={self.__dict__})")

# # Ví dụ sử dụng lớp
# list_item_button_props = ListItemButtonProps(
#     component=str,
#     additional_props={'prop1': 'value1'},
#     default_component='div',
#     alignItems='center',
#     autoFocus=True
# )

# print(list_item_button_props)




from typing import Dict, Callable

# Giả lập đối tượng CSSObject, đại diện cho các thuộc tính CSS
class CSSObject:
    css: str


class ApplyStyles:
    """
    Class ApplyStyles để áp dụng style dựa trên chế độ màu từ theme.
    """
    def __init__(self, theme: Dict[str, Callable[[str], str]]):
        """
        :param theme: Đối tượng theme chứa các phương thức hoặc biến liên quan đến chế độ màu.
        """
        self.theme = theme

    def __call__(self, key: str, styles: CSSObject) -> CSSObject:
        """
        Áp dụng style dựa trên key (chế độ màu) và trả về các thuộc tính CSS tương ứng.
        
        :param key: Chế độ màu (ví dụ: "light", "dark")
        :param styles: Các thuộc tính CSS cần áp dụng
        :return: Các style đã được áp dụng nếu khớp với chế độ màu hoặc một đối tượng trống
        """
        if self.theme.get('vars') and callable(self.theme.get('getColorSchemeSelector')):
            # Lấy selector dựa trên key (ví dụ: 'light' hoặc 'dark')
            selector = self.theme['getColorSchemeSelector'](key).replace(r'(\[[^\]]+\])', '*:where($1)')
            return {selector: styles}
        
        # Nếu chế độ màu của palette khớp với key
        if self.theme.get('palette', {}).get('mode') == key:
            return styles

        # Nếu không khớp, trả về đối tượng trống
        return {}

# # Ví dụ sử dụng ApplyStyles
# theme = {
#     'palette': {'mode': 'dark'},
#     'vars': None,
#     'getColorSchemeSelector': lambda key: f'[data-mui-color-scheme="{key}"]'
# }

# # Khởi tạo ApplyStyles với theme
# apply_styles = ApplyStyles(theme)

# # Áp dụng các style cho chế độ 'dark'
# styles = apply_styles('dark', {'background': '#1c1c1c', 'color': '#fff'})
# print(styles)  # Kết quả: {'[data-mui-color-scheme="dark"]': {'background': '#1c1c1c', 'color': '#fff'}}




# Copyright (C) 2022 The Pyact Company Ltd.
# SPDX-License-Identifier: LicenseRef-Pyact-Commercial OR BSD-3-Clause
from __future__ import annotations

"""..site_packages.qtcompat port of the areachart example from qtmui v1.0"""

from typing import Optional, Union
import sys
from PySide6.QtCore import QPointF, Qt
from PySide6.QtWidgets import QMainWindow, QApplication, QWidget, QVBoxLayout
from PySide6.QtCharts import (QChart, 
                              QChartView, 
                              QLineSeries, 
                              QAreaSeries,
                              QBarCategoryAxis, 
                              QBarSeries, 
                              QBarSet, 
                              QChart,
                              QChartView, 
                              QValueAxis,
                              )
from PySide6.QtGui import QGradient, QPen, QLinearGradient, QPainter

from .map_change_theme import MapChangeTheme


class Map(QWidget):
    def __init__(
                self,
                initialViewState: dict = None, 
                *args,
                **kwargs
                ):
        super().__init__(*args, **kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0,0,0,0)

        map = MapChangeTheme(
            initialViewState=initialViewState,
        )
        self.layout().addWidget(map)
